# Rebuild: Liu Pei  <liupei200546@163.com>  2022/03/18

from itertools import chain
from typing import Optional

import kernel
import numpy as np

from .tools import generate_spanlist

__all__ = [
    'ZZRamsey',
    'CR_tomo_amp1', 'CR_tomo_others', 'CR_seq', 'CR_seq_param',
]


def _ZZ_circuit(qubit: str, neighbors: list[str],
                delay: float = 50e-9, **kw) -> list:
    """
    Generate ZZ initial circuits.

    Args:
        qubit (str): qubit observed.
        neighbors (list[str]): neighbors of the qubit.
        delay (float, optional): delay time. Defaults to 50e-9.

    Returns:
        list: initial circuits.
    """
    ret = []
    for i in range(2**(len(neighbors))):
        tmp = []
        for j in range(len(neighbors)):
            if i & (2**j):
                tmp.append(('X', neighbors[j]))
            else:
                tmp.append(('I', neighbors[j]))
        tmp.extend([('Barrier', tuple([qubit, *neighbors])),
                    (('Delay', delay), qubit),
                    ('Barrier', tuple([qubit, *neighbors]))])
        ret.append(tmp)
    return ret


def ZZRamsey(qubit: str, neighbors: list[str],
             rotate: float, time_length: float, points_in_one_period: int = 10,
             signal: str = 'population', **kw) -> dict:
    """
    ['time'] ZZ using Ramsey to watch.

    Args:
        qubit (str): qubit observed.
        neighbors (list[str]): neighbors of the qubit.
        rotate (float): offset of frequency.
        time_length (float): length of delay time.
        points_in_one_period (int, optional): points per period. Defaults to 10.
        signal (str, optional): signal. Defaults to 'population'.
    """

    from waveforms.quantum import Ramsey

    return {
        'init': {
            'name': 'ZZ Ramsey',
            'qubits': [qubit, *neighbors],
            'signal': signal,
        },
        'sweep_setting': {
            'gate': _ZZ_circuit(qubit=qubit, neighbors=neighbors, **kw),
            'time': np.linspace(0, time_length, round(time_length*rotate*points_in_one_period)+1),
        },
        'sweep_addition': {
            'circuit': lambda gate, time: [
                *gate,
                *Ramsey(qubit=qubit, t=time, f=rotate),
                (('Measure', 0), qubit),
            ],
        },
    }


def _cr_readout_list(qubits: list[tuple[str]],
                     readout: str = 'target',
                     **kw) -> list[str]:

    if readout == 'both':
        readout_qubits = chain.from_iterable(qubits)
    elif readout == 'target':
        readout_qubits = [q[1] for q in qubits]
    elif readout == 'control':
        readout_qubits = [q[0] for q in qubits]
    else:
        raise ValueError('Illegal readout type in generate a cr circuit')

    return readout_qubits


def CR_tomo_amp1(qubits: list[tuple[str]],
                 st: Optional[float] = None, ed: Optional[float] = None, sweep_points: int = 13, mode: str = 'linear',
                 tomo_time_length: float = 10, tomo_time_step: float = 0.1,
                 gate1: list = ['I', 'X'], gate2: list = ['-Y/2', 'X/2', 'I'],
                 signal: str = 'population', default_type: str = 'default', **kw) -> dict:
    """
    [Feed 'duration'] H tomography for CR as amp1 changing.

    Args:
        qubits (list[tuple[str]]): qubit names.
        st (Optional[float], optional): scale start. Defaults to None.
        ed (Optional[float], optional): scale end. Defaults to None.
        sweep_points (int, optional): sweep points. Defaults to 13.
        mode (str, optional): sampling mode. Defaults to 'linear'.
        tomo_time_length (float, optional): times of duration. Defaults to 10.
        tomo_time_step (float, optional): step of duration. Defaults to 0.1.
        gate1 (list, optional): gate on control. Defaults to ['I', 'X'].
        gate2 (list, optional): gate on target. Defaults to ['-Y/2', 'X/2', 'I'].
        signal (str, optional): signal. Defaults to 'population'.
    """

    config_type = 'params' if default_type == 'default' else default_type

    amp1 = {
        q: kernel.get(f'gate.CR.{q[0]}_{q[1]}.{default_type}.amp1') for q in qubits
    }
    strength = {
        q: kernel.get(f'gate.CR.{q[0]}_{q[1]}.{default_type}.duration')*amp1[q] for q in qubits
    }
    readout_qubits = _cr_readout_list(qubits=qubits, **kw)

    return {
        'init': {
            'name': 'CR Tomo Amp1',
            'qubits': qubits,
            'signal': signal,
        },
        'setting': {
            'feed': lambda _, step, amp1=amp1, strength=strength, qubits=qubits: step.feed(
                {
                    'duration': [
                        strength[q]/amp1[q]/step.kwds['amp1_times']*step.kwds['duration_times'] for q in qubits
                    ]
                },
                store=True),
        },
        'sweep_setting': {
            'amp1_times': generate_spanlist(st=st, ed=ed, sweep_points=sweep_points, mode=mode),
            'duration_times': np.arange(tomo_time_step, tomo_time_length + tomo_time_step, tomo_time_step),
            'gate1': gate1,
            'gate2': gate2,
        },
        'sweep_addition': {
            'circuit':
            lambda amp1_times, duration_times, gate1, gate2, amp1=amp1, strength=strength, qubits=qubits: [
                *[(gate1, q[0]) for q in qubits],
                *[(('CR', ('with',
                           ('type', default_type), 
                           ('param:duration',
                            strength[q]/amp1[q]/amp1_times*duration_times),
                           ('param:amp1', amp1[q]*amp1_times))), q) for q in qubits],
                *[(gate2, q[1]) for q in qubits],
                ('Barrier', tuple(chain.from_iterable(qubits))),
                *[(('Measure', j), q) for j, q in enumerate(readout_qubits)],
            ],
        },
    }


def CR_tomo_others(qubits: list[tuple[str]], name: str,
                   st: Optional[float] = None, ed: Optional[float] = None, mode: str = 'linear', sweep_points: int = 13,
                   tomo_time_length: float = 10, tomo_time_step: float = 0.1, duration: Optional[dict] = None,
                   gate1: list = ['I', 'X'], gate2: list = ['-Y/2', 'X/2', 'I'],
                   signal: str = 'population', multi_param: list[str] = ['amp2'], **kw) -> dict:
    """
    [Feed 'duration'] H tomography for CR as amp1 changing.

    Args:
        qubits (list[tuple[str]]): qubit names.
        name (str): name of params.
        st (Optional[float], optional): scale start. Defaults to None.
        ed (Optional[float], optional): scale end. Defaults to None.
        sweep_points (int, optional): sweep points. Defaults to 13.
        mode (str, optional): sampling mode. Defaults to 'linear'.
        tomo_time_length (float, optional): times of duration. Defaults to 10.
        tomo_time_step (float, optional): step of duration. Defaults to 0.1.
        gate1 (list, optional): gate on control. Defaults to ['I', 'X'].
        gate2 (list, optional): gate on target. Defaults to ['-Y/2', 'X/2', 'I'].
        signal (str, optional): signal. Defaults to 'population'.
        multi_param (list[str], optional): params not in `multi_param` using add mode. Defaults to ['amp2'].
    """

    sweep_list = generate_spanlist(
        st=st, ed=ed, sweep_points=sweep_points, mode=mode)

    if duration is None:
        duration = {
            q: kernel.get(f'gate.CR.{q[0]}_{q[1]}.params.duration')
            for q in qubits
        }
    para_list = []
    for q in qubits:
        if name in multi_param:
            para_list.append(
                sweep_list*kernel.get(f'gate.CR.{q[0]}_{q[1]}.params.{name}'))
        else:
            para_list.append(
                sweep_list+kernel.get(f'gate.CR.{q[0]}_{q[1]}.params.{name}'))
    readout_qubits = _cr_readout_list(qubits=qubits, **kw)

    return {
        'init': {
            'name': f'CR Tomo {name}',
            'qubits': qubits,
            'signal': signal,
        },
        'setting': {
            'feed': lambda _, step, duration=duration, qubits=qubits: step.feed(
                {
                    'duration': [
                        step.kwds['duration_times'] * duration[q]
                        for q in qubits
                    ]
                },
                store=True),
        },
        'sweep_setting': {
            tuple([
                f'{q[0]}_{q[1]}' for q in qubits
            ]): tuple(para_list),
            'duration_times': np.arange(tomo_time_step, tomo_time_length + tomo_time_step, tomo_time_step),
            'gate1': gate1,
            'gate2': gate2,
        },
        'sweep_addition': {
            'circuit': lambda duration_times, gate1, gate2, duration=duration, qubits=qubits, name=name, **kw: [
                *[(gate1, q[0]) for q in qubits],
                *[(('CR', ('with',
                           ('param:duration', duration[q] * duration_times),
                           (f'param:{name}', kw[f'{q[0]}_{q[1]}']))), q) for q in qubits],
                *[(gate2, q[1]) for q in qubits],
                ('Barrier', tuple(chain.from_iterable(qubits))),
                *[(('Measure', j), q) for j, q in enumerate(readout_qubits)],
            ],
        },
    }


def _cr_seq_circuit(qubits: list[tuple[str]],
                    seq: str = 'A', readout: str = 'target',
                    n: int = 1) -> dict:
    """
    Generate CR gates. Ref arXiv: 2106.00675.

    Args:
        qubits (list[tuple[str]]): qubits name.
        seq (str, optional): sequency type. Defaults to 'A'.
        readout (str, optional): readout type. Defaults to 'target'.
        n (int, optional): repeat number. Defaults to 1.

    Raises:
        ValueError: 'Illegal sequency type in generate a cr circuit'.
    """

    readout_qubits = _cr_readout_list(qubits=qubits, readout=readout)

    ret = []
    if seq == 'A':
        ret.extend([
            *[('X/2', q[1]) for q in qubits], *[('CR', q) for q in qubits] * n
        ])
    elif seq == 'B':
        ret.extend([
            *[('X', q[0]) for q in qubits], *[('X/2', q[1]) for q in qubits],
            *[('CR', q) for q in qubits] * n
        ])
    elif seq == 'C':
        ret.extend([
            *[('X/2', q[1]) for q in qubits], *[('CR', q) for q in qubits] * n,
            *[('Y/2', q[1]) for q in qubits]
        ])
    elif seq == 'D':
        ret.extend([
            *[('X', q[0]) for q in qubits],
            *chain.from_iterable(
                zip([*[('CR', q) for q in qubits] * n],
                    [*[(('CR', ('with', 
                    ('param:global_phase', kernel.get(f'gate.CR.{q[0]}_{q[1]}.params.global_phase')+np.pi)))
                    , q) for q in qubits] * n])),
            *[('Y/2', q[1]) for q in qubits],
        ])
    elif seq == 'E':
        ret.extend([
            *[('Y/2', q[1]) for q in qubits], *[('CR', q) for q in qubits] * n
        ])
    elif seq == 'F':
        ret.extend([
            *[('X', q[0]) for q in qubits], *[('X/2', q[1]) for q in qubits],
            *chain.from_iterable(
                zip([*[('-Y', q[1]) for q in qubits] * n],
                    [*[('CR', q) for q in qubits] * n])),
            *[('-Y/2', q[1]) for q in qubits]
        ])
    elif seq == 'G':
        ret.extend([
            *[('X/2', q[0]) for q in qubits],
            *[('CR', q) for q in qubits] * n * 2,
            *[(('Rz', n * np.pi / 2), q[0]) for q in qubits], *[('X/2', q[0])
                                                                for q in qubits]
        ])
    else:
        raise ValueError('Illegal sequency type in generate a cr circuit')

    ret.extend([('Barrier', tuple(chain.from_iterable(qubits))),
                *[(('Measure', j), q) for j, q in enumerate(readout_qubits)]])

    return ret


def CR_seq(qubits: list[tuple[str]],
           seq: str = 'A', readout: Optional[str] = None,
           max_n: int = 10, step_n: float = 1,
           signal: str = 'population', **kw) -> dict:
    """
    ['n'] Repeat N in circuit.

    Args:
        qubits (list[tuple[str]]): qubits name.
        seq (str, optional): sequency type. Defaults to 'A'.
        readout (Optional[str], optional): readout type. Defaults to None.
        max_n (int, optional): maximum of n. Defaults to 10.
        step_n (float, optional): step of n. Defaults to 1.
        signal (str, optional): signal. Defaults to 'population'.
    """

    assert seq in ['A', 'B', 'C', 'D', 'E', 'F',
                   'G'], 'Illegal sequency type in generate a cr circuit'

    if readout is None:
        readout = 'control' if seq == 'G' else 'target'
    else:
        assert readout in ['both', 'target', 'control'
                           ], 'Illegal readout type in generate a cr circuit'

    return {
        'init': {
            'name': f'CR Seq {seq}',
            'qubits': qubits,
            'signal': signal,
        },
        'sweep_setting': {
            'n': np.arange(step_n, max_n + 1, step_n),
        },
        'sweep_addition': {
            'circuit':
            lambda n: _cr_seq_circuit(
                qubits=qubits, readout=readout, seq=seq, n=n),
        },
    }


def CR_seq_param(qubits: list[tuple[str]], name: str,
                 seq: str = 'A', readout: Optional[str] = 'target',
                 max_n: int = 10, step_n: float = 1,
                 center: Optional[float] = None, delta: Optional[float] = None, st: Optional[float] = None, ed: Optional[float] = None, sweep_points: int = 101, mode: str = 'linear',
                 signal: str = 'population', multi_param: list[str] = ['amp1', 'amp2'], default_type: str = 'params', **kw) -> dict:
    """

    Args:
        qubits (list[tuple[str]]): qubits name.
        name (str): params name.
        seq (str, optional): sequency type. Defaults to 'A'.
        readout (Optional[str], optional): readout type. Defaults to None.
        max_n (int, optional): maximum of n. Defaults to 10.
        step_n (float, optional): step of n. Defaults to 1.
        center (Optional[float], optional): sweep center. Defaults to None.
        delta (Optional[float], optional): sweep span. Defaults to None.
        st (Optional[float], optional): sweep start. Defaults to None.
        ed (Optional[float], optional): sweep end. Defaults to None.
        sweep_points (int, optional): sweep points. Defaults to 101.
        mode (str, optional): sweep mode. Defaults to 'linear'.
        signal (str, optional): signal. Defaults to 'population'.
        multi_param (list[str], optional): params not in `multi_param` using add mode. Defaults to ['amp2'].
    """

    assert seq in ['A', 'B', 'C', 'D', 'E', 'F',
                   'G'], 'Illegal sequency type in generate a cr circuit'

    if readout is None:
        readout = 'control' if seq == 'G' else 'target'
    else:
        assert readout in ['both', 'target', 'control'
                           ], 'Illegal readout type in generate a cr circuit'

    sweep_list = generate_spanlist(
        center=center, delta=delta, st=st, ed=ed, sweep_points=sweep_points, mode=mode)
    para_list = []
    for q in qubits:
        if name in multi_param:
            para_list.append(
                sweep_list*kernel.get(f'gate.CR.{q[0]}_{q[1]}.{default_type}.{name}'))
        else:
            para_list.append(
                sweep_list+kernel.get(f'gate.CR.{q[0]}_{q[1]}.{default_type}.{name}'))

    return {
        'init': {
            'name': f'CR Seq {seq} {name}',
            'qubits': qubits,
            'signal': signal,
        },
        'sweep_config': {
            f'{q[0]}_{q[1]}': {'addr': f'gate.CR.{q[0]}_{q[1]}.{default_type}.{name}'} for q in qubits
        },
        'sweep_setting': {
            tuple([
                f'{q[0]}_{q[1]}' for q in qubits
            ]): tuple(para_list),
            'n': np.arange(step_n, max_n + 1, step_n),
        },
        'sweep_addition': {
            'circuit': lambda n: _cr_seq_circuit(qubits=qubits, readout=readout, seq=seq, n=n),
        },
    }



# Author: lizhiyuan  <lizhiyuan@baqis.ac.cn>  2021/01/29

import numpy as np
from qos_tools.fit.fitclass import FitClass
from qos_tools.fit import function, plotscript

__all__ = [
    'T1', 'Rabi', 'Ramsey', 'Ramsey_withBeat', 'SpinEcho', 'RB'
]

def T1(data, fixed=None, p0=None, **kw):
    fit = FitClass(data, function._T1, fixed=fixed, p0=p0, **kw)
    fit._plotscript = plotscript.T1
    return fit


def Rabi(data, fixed=None, p0=None, **kw):
    fit = FitClass(data, function._Rabi, fixed=fixed, p0=p0, **kw)
    fit.params['pplen'] = np.abs(0.5/fit.params['Omega'])
    # fit.error['pplen'] = None
    fit._plotscript = plotscript.Rabi
    return fit


def Ramsey(data, fixed=None, p0=None, **kw):
    fit = FitClass(data, function._Ramsey, fixed=fixed, p0=p0, **kw)
    fit.params['Tphi'] = abs(fit.params['Tphi'])
    fit.params['T2star'] = 1/(0.5/fit.params['T1'] + 1/fit.params['Tphi'])
    # fit.error['T2star'] = None
    fit._plotscript = plotscript.Ramsey
    return fit


def Ramsey_withBeat(data, fixed=None, p0=None, **kw):
    fit = FitClass(data, function._Ramsey_withBeat, fixed=fixed, p0=p0, **kw)
    fit.params['Tphi'] = abs(fit.params['Tphi'])
    fit.params['T2star'] = 1/(0.5/fit.params['T1'] + 1/fit.params['Tphi'])
    # fit.error['T2star'] = None
    fit._plotscript = plotscript.Ramsey_withBeat
    return fit


def SpinEcho(data, fixed=None, p0=None, **kw):
    fit = FitClass(data, function._Ramsey, fixed=fixed, p0=p0, **kw)
    fit.params['Tphi'] = abs(fit.params['Tphi'])
    fit.params['T2E'] = 1/(0.5/fit.params['T1'] + 1/fit.params['Tphi'])
    # fit.error['T2E'] = None
    fit._plotscript = plotscript.SpinEcho
    return fit


def RB(data, fixed=None, p0=None, d=2, **kw):
    '''d: d-dimensional system, for the Clifford group, d=2'''
    fit = FitClass(data, function._RB, fixed=fixed, p0=p0, **kw)
    p, p_e = fit.params['p'], fit.error['p']
    fit.params['Fidelity'] = 1-(1-p)*(d-1)/d
    # fit.error['Fidelity'] = p_e*(1-d)/d
    fit._plotscript = plotscript.RB
    return fit



# Author: Liu Pei  <liupei200546@163.com>  2022/07/25

from dataclasses import dataclass
from queue import Queue
from typing import Any, Sequence, Union

import nevergrad as ng
from scipy.optimize import OptimizeResult
from waveforms.scan_iter import BaseOptimizer


@dataclass
class VarData:
    pattern: str = 'mul'
    init: float = 1
    low: float = 0.9
    high: float = 1.1

    def get(self, key, default=None):
        return self.__getattribute__(key)


def init_variable(variable: list[dict]) -> ng.p.Array:
    init, low, high = [], [], []
    for item in variable:
        init.append(item.get('init', 1))
        if item.get('pattern', 'mul') in ['add']:
            low.append(init[-1] + item.get('low', 0.9))
            high.append(init[-1] + item.get('high', 1.1))
        else:
            low.append(init[-1] * item.get('low', 0.9))
            high.append(init[-1] * item.get('high', 1.1))
    return ng.p.Array(init=init, lower=low, upper=high)


class NgOpt(BaseOptimizer):

    def __init__(self,
                 variable: Union[list[dict], list[VarData], ng.p.Array],
                 config: dict = {},
                 **kw):

        self._catch, self._all_x, self._all_y = Queue(), [], []
        self.config = {
            'method': 'TBPSA',
            'budget': 300,
        }

        self.config.update(config)
        if not isinstance(variable, ng.p.Array):
            variable = init_variable(variable=variable)
        self.opt = getattr(ng.optimizers,
                           self.config['method'])(variable,
                                                  budget=self.config['budget'])

    def ask(self):
        tmp = self.opt.ask()
        self._catch.put(tmp)
        return tmp.args[0]

    def tell(self, suggested: Sequence, value: Any):
        if not self._catch.empty():
            tmp = self._catch.get()
            self._all_x.append(tmp.args[0])
            self._all_y.append(value)
            self.opt.tell(tmp, value)

    def get_result(self, history: bool = False) -> OptimizeResult:
        ret = OptimizeResult({'x': self.opt.recommend().args[0]})
        if history:
            ret.x_iters = self._all_x
            ret.func_vals = self._all_y
        return ret



# Author: lizhiyuan  <lizhiyuan@baqis.ac.cn>  2021/01/01

from ._equal_utils import is_equal
from ._dict_utils import flatten_dict
from blinker import Signal
import asyncio
import logging
log = logging.getLogger(__name__)


class ConsManager(object):
    '''
    约束管理器

    按照给定的约束规则，管理一个容器内部的约束关联
    '''

    def __init__(self, container, constrains=[], init=True, delimiter='.'):
        '''
        Parameters:
            container: object
                一个类字典的容器，支持方括号索引的形式设置和读取，或支持set/get方法

            constrains: list
                约束条目，每个元素都为一个三元元组（deps, func, target），例如，
                [
                    (('k1.k11','k1.k12'),   (lambda v1,v2:v1+v2),   'k2'       ),
                    (('k3','k4'),           (lambda v1,v2:(v1,v2)), ('k5','k6')),
                ]

            init: bool, optional
                是否使用约束条件初始化传入的字典，默认 True

            delimiter: str, optional
                分割符，默认 .
        '''
        self.container = container
        self.delimiter = delimiter
        self.sig = Signal()
        # 如果约束的 func 是字符串，则 eval
        _constrains = []
        for deps, func, target in constrains:
            if isinstance(func, str):
                func = eval(func)
            _constrains.append((deps, func, target))

        self.constrains = _constrains
        self.key_changed = set()  # 集合，记录改变的key，留出接口
        self.subscribers = self.__init_cons(_constrains, init=init)

    def __init_cons(self, constrains, init=True):
        subscribers = []
        for deps, func, target in constrains:
            subs = _subscrib(deps, func, target, self)
            subscribers.append(subs)
            for _k in deps:
                self.sig.connect(subs, sender=_k)
        if init:
            deps_set = set()
            for deps, func, target in constrains:
                deps_set.update(deps)
            for dep in deps_set:
                self._send(dep)
        return subscribers

    def _send(self, key):
        self.sig.send(key)
        self.key_changed.add(key)
        _k_list = key.rsplit(self.delimiter, maxsplit=1)
        if len(_k_list) > 1:
            self._send(_k_list[0])

    def _set(self, key, value, check=True):
        '''根据关键词连接的字符串，设置对应的值

        Parameters:
            key: str or list
                使用符号串联关键词的字符串，比如'pulse.width.value'

            value: 
                待设入的值

            check: bool
                是否检查设入值与原值相等
        Return: 
            bool, True 表示value改变并且已设置，False 表示未变
        '''
        ks = key.rsplit(self.delimiter,maxsplit=1)
        if len(ks) > 1:
            _ks, _k = ks[0], ks[1]
            _ct = self.get(_ks)
        else:
            _k = key
            _ct = self.container
        assert _ct is not None, f"key '{key}' not exist !"

        if check and is_equal(value, _ct.get(_k)):
            return False
        else:
            try:
                _ct[_k] = value
            except TypeError:  # object is not subscriptable
                _ct.set(_k, value)
            except Exception as e:
                raise e
            return True
    
    def set(self, key, value, check=True):
        is_changed = self._set(key, value, check=check)
        if is_changed:
            self._send(key)

    def get(self, key, default=None):
        '''根据关键词连接的字符串或者列表，获取对应的值

        Parameters:
            key: str or list
                使用符号串联关键词的字符串，比如'pulse.width.value'

            default: 
                类似于字典方法，当关键词不存在时返回一个默认值

        Return:
            返回键值串对应的值
        '''
        ks = key.split(self.delimiter)
        _v = self.container
        for k in ks:
            try:
                _v = _v[k]
            except KeyError:
                _v = default
            except TypeError:  # object is not subscriptable
                _v = _v.get(k)
            except Exception as e:
                raise e
        return _v

    def update(self, d, check=True):
        '''批量更新信息

        Parameters:
            d: dict
                待更新信息的字典

            check: bool,optional
                是否检查设入值与原值相等
        '''
        k_changed = []
        d_flatten = flatten_dict(d, self.delimiter)
        for k, v in d_flatten.items():
            _changed = self._set(k, v, check=check)
            if _changed:
                k_changed.append(k)
        for _k in k_changed:
            self._send(_k)


def _subscrib(deps, func, target, cons_manager):
    '''订阅函数生成器

    根据需要的参量产生一个订阅函数

    Parameters:
        deps,func,target: 
            依赖关键词[列表或元组]，约束函数，目标关键词[单key(字符串)，或者多key(列表/元组)]，
            比如 (('k1','k2'), lambda v1,v2:v1+v2, 'k3'), func的返回值应与target数目一致

        cons_manager: object
            ConsManager的一个实例

    Return:
        返回一个订阅函数
    '''
    def subscriber(send):
        arg = [cons_manager.get(k_dep) for k_dep in deps]
        if None not in arg:
            if asyncio.iscoroutinefunction(func):
                loop = asyncio.get_event_loop()
                task = func(*arg)
                value = loop.run_until_complete(task)
            else:
                value = func(*arg)
            if isinstance(target, str):
                cons_manager.set(target, value)
            else:
                d = dict(zip(target, value))
                cons_manager.update(d)
        else:  # 这里只打出警告：一些依赖变量的值为None，打印出所有的键值对
            log.warning(f'{__name__} WARNING: ' +
                        ', '.join([f'{k} is {v}' for k, v in zip(deps, arg)]))
    return subscriber



# Author: Liu Pei  <liupei200546@163.com>  2021/08/14

from itertools import chain
from typing import Optional, Union

import kernel
import numpy as np

__all__ = [
    'S21_base_change_lo_with_constrains', 'S21_base_change_lo_without_constrains', 'S21_base_change_awg', 'S21_change_lo_without_constrains',
    'Spectrum_base_change_awg_without_constrains', 'Spectrum_base_change_lo_without_constrains', 'Spectrum_base_change_lo_without_mixing_wave'
    'PowerRabi_base_scaleTimes', 'PowerRabi_base_n_pulse', 'TimeRabi_base_durationTimes',
    'T1_without_constrains',
    'Ramsey_without_constrains', 'SpinEcho_without_constrains', 'CPMG_without_constrains', 'CP_without_constrains',
    'AllXY_without_constrains', 'AllXY6and16_alpha_without_constrains', 'AllXY7and8_delta_without_constrains', 'AllXY11and12_beta_without_constrains',
    'ReadoutFrequency_without_constrains', 'ReadoutAmp_without_constrains', 'Scatter2',
    'ReadoutDelay_without_constrains', 'RTO_without_constrains', 'PiErrorOscillation_without_constrain', 'DRAGCheck_without_constrain',
    'RB_single_qubit_without_constrain', 'XEB_single_qubit_without_constrain', 'nop',
    'XXDelta_without_constrain',
]


def generate_spanlist(center: Optional[float] = None, delta: Optional[float] = None, st: Optional[float] = None, ed: Optional[float] = None,
                      sweep_points: int = 101) -> np.ndarray:

    if center is not None and delta is not None:
        return np.linspace(center-delta, center+delta, sweep_points)
    elif st is not None and ed is not None:
        return np.linspace(st, ed, sweep_points)
    else:
        raise ValueError('The scanning interval entered is illegal')


def generate_spanlist_log(center: Optional[float] = None, delta: Optional[float] = None, st: Optional[float] = None, ed: Optional[float] = None,
                          sweep_points: int = 101) -> np.ndarray:
    sweep_half = (np.logspace(0, 1, (sweep_points+1)//2)-1)/9
    if center is not None and delta is not None:
        return np.concatenate((-sweep_half[::-1][:-1], sweep_half))*delta+center
    elif st is not None and ed is not None:
        if center is None:
            center = (st+ed)/2
            delta = (ed-st)/2
            return np.concatenate((-sweep_half[::-1][:-1], sweep_half))*delta+center
        else:
            return np.concatenate((sweep_half[::-1][:-1]*(st-center), sweep_half*(ed-center)))+center
    else:
        raise ValueError('The scanning interval entered is illegal')


def generate_RB_cycle(max_cycle: int, sweep_points: int, start: float = 0):

    return np.unique(np.round(np.logspace(start, np.log10(max_cycle*1.0), sweep_points))).astype(int)


def S21_base_change_lo_with_constrains(qubits: list[int], delta: Optional[float] = None, st: Optional[float] = None, ed: Optional[float] = None,
                                       sweep_points: int = 101) -> dict:
    """
    ['delta'] Measure S21 with constraints, change local frequency.

    Args:
        qubits (list[int]): the qubit id.
        delta (Optional[float], optional): span frequency, in the unit Hz. The sweep frequency is (-delta, delta) plus measure frequency read from config. Defaults to None.
        st (Optional[float], optional): span frequency start, in the unit Hz. The sweep frequency is (st, ed) plus measure frequency read from config. Defaults to None.
        ed (Optional[float], optional): span frequency end, in the unit Hz. The sweep frequency is (st, ed) plus measure frequency read from config. Defaults to None. Defaults to None.
        sweep_points (int, optional): sweep points. Defaults to 101.

    Returns:
        dict: config.
    """

    sweep_list = generate_spanlist(
        center=0, delta=delta, st=st, ed=ed, sweep_points=sweep_points)
    measures = set(kernel.get(f'Q{i}.probe')for i in qubits)

    return {
        'init': {
            'name': 'S21',
            'qubits': qubits,
            'signal': 'raw',
            'scanner_name': 'S21_base_change_lo_with_constrains',
        },
        'setting': {
            'delta': 0,
            'qubit_frequency': {f'Q{i}': kernel.get(f'gate.Measure.Q{i}.params.frequency') for i in qubits},
            'LO_frequency': {f'{m}': kernel.get(f'{m}.setting.LO') for m in measures},
            'pre_setting': {},
            'compile_once': True,
            'circuit_type': 'cmds:qlispcmd',
            'circuit': [
                (('Measure', j), f'Q{i}') for j, i in enumerate(qubits)
            ],
        },
        'constrains': [
            *[((f'setting.qubit_frequency.Q{i}', 'setting.delta'), lambda a,
               b: a+b, f'gate.Measure.Q{i}.params.frequency') for i in qubits],
            *[((f'setting.LO_frequency.{m}', 'setting.delta'),
               lambda a, b: a+b, f'{m}.setting.LO') for m in measures],
        ],
        'sweep_config': {
            'delta': {
                'addr': 'setting.delta',
            },
        },
        'sweep_setting': {
            'delta': sweep_list,
        },
        'plot_setting': [
            ['Frequency', np.asarray([sweep_list for _ in qubits]), 'Hz'],
        ],
        'plot': [
            ['Frequency', 'Hz'], ['Amplitude', 'a.u.'],
        ],
    }


def S21_base_change_lo_without_constrains(qubits: list[int], delta: Optional[float] = None, st: Optional[float] = None, ed: Optional[float] = None,
                                          sweep_points: int = 101) -> dict:
    """
    [f'Q{i}'] Measure S21 without constraints, change local frequency.

    Args:
        qubits (list[int]): the qubit id.
        delta (Optional[float], optional): span frequency, in the unit Hz. The sweep frequency is (-delta, delta) plus measure frequency read from config. Defaults to None.
        st (Optional[float], optional): span frequency start, in the unit Hz. The sweep frequency is (st, ed) plus measure frequency read from config. Defaults to None.
        ed (Optional[float], optional): span frequency end, in the unit Hz. The sweep frequency is (st, ed) plus measure frequency read from config. Defaults to None. Defaults to None.
        sweep_points (int, optional): sweep points. Defaults to 101.

    Returns:
        dict: cogfig.
    """

    sweep_list = generate_spanlist(
        center=0, delta=delta, st=st, ed=ed, sweep_points=sweep_points)
    measures = set(kernel.get(f'Q{i}.probe')for i in qubits)

    return {
        'init': {
            'name': 'S21',
            'qubits': qubits,
            'signal': 'raw',
            'scanner_name': 'S21_base_change_lo_without_constrains',
        },
        'setting': {
            'pre_setting': {},
            'compile_once': True,
            'circuit_type': 'cmds:qlispcmd',
            'circuit': [
                (('Measure', j), f'Q{i}') for j, i in enumerate(qubits)
            ],
        },
        'constrains': [],
        'sweep_config': dict(
            list({
                f'Q{i}': {
                    'addr': f'gate.Measure.Q{i}.params.frequency',
                } for i in qubits
            }.items())+list({
                m: {
                    'addr': f'{m}.setting.LO',
                } for m in measures
            }.items())
        ),
        'sweep_setting': {
            tuple([
                *[f'Q{i}' for i in qubits],
                *[m for m in measures],
            ]): tuple([
                *[sweep_list+kernel.get(f'gate.Measure.Q{i}.params.frequency') for i in qubits],
                *[sweep_list+kernel.get(f'{m}.setting.LO') for m in measures],
            ]),
            
        },
        'plot_setting': [
            ['Frequency', np.asarray(
                [sweep_list+kernel.get(f'gate.Measure.Q{i}.params.frequency') for i in qubits]), 'Hz'],
        ],
        'plot': [
            ['Frequency', 'Hz'], ['Amplitude', 'a.u.'],
        ],
    }


def S21_base_change_awg(qubits: list[int], delta: Optional[float] = None, st: Optional[float] = None, ed: Optional[float] = None,
                        sweep_points: int = 101) -> dict:
    """
    [f'Q{i}'] Measure S21 without constraints, change awg frequency.

    Args:
        qubits (list[int]): the qubit id.
        delta (Optional[float], optional): span frequency, in the unit Hz. The sweep frequency is (-delta, delta) plus measure frequency read from config. Defaults to None.
        st (Optional[float], optional): span frequency start, in the unit Hz. The sweep frequency is (st, ed) plus measure frequency read from config. Defaults to None.
        ed (Optional[float], optional): span frequency end, in the unit Hz. The sweep frequency is (st, ed) plus measure frequency read from config. Defaults to None. Defaults to None.
        sweep_points (int, optional): sweep points. Defaults to 101.

    Returns:
        dict: config.
    """

    sweep_list = generate_spanlist(
        center=0, delta=delta, st=st, ed=ed, sweep_points=sweep_points)

    return {
        'init': {
            'name': 'S21',
            'qubits': qubits,
            'signal': 'raw',
            'scanner_name': 'S21_base_change_awg',
        },
        'setting': {
            'pre_setting': {},
            'compile_once': False,
            'circuit_type': 'cmds:qlispcmd',
            'circuit': [
                (('Measure', j), f'Q{i}') for j, i in enumerate(qubits)
            ],
        },
        'constrains': [],
        'sweep_config': {
            f'Q{i}': {
                'addr': f'gate.Measure.Q{i}.params.frequency',
            } for i in qubits
        },
        'sweep_setting': {
            tuple([
                f'Q{i}' for i in qubits
            ]): tuple([
                sweep_list+kernel.get(f'gate.Measure.Q{i}.params.frequency') for i in qubits
            ]),
        },
        'plot_setting': [
            ['Frequency', np.asarray(
                [sweep_list+kernel.get(f'gate.Measure.Q{i}.params.frequency') for i in qubits]), 'Hz'],
        ],
        'plot': [
            ['Frequency', 'Hz'], ['Amplitude', 'a.u.'],
        ],
    }


def S21_change_lo_without_constrains(qubits: list[int], center: Optional[float] = None, delta: Optional[float] = None, st: Optional[float] = None,
                                     ed: Optional[float] = None, sweep_points: int = 101, sideband: Optional[float] = None) -> dict:
    """
    [f'Q{i}'] Measure S21 without constraints, change lo frequency.

    Args:
        qubits (list[int]): the qubit id.
        center (Optional[float], optional): center frequency, in the unit Hz. Defaults to None.
        delta (Optional[float], optional): span frequency, in the unit Hz. The sweep frequency is (-delta, delta) plus measure frequency read from config. Defaults to None.
        st (Optional[float], optional): span frequency start, in the unit Hz. The sweep frequency is (st, ed) plus measure frequency read from config. Defaults to None.
        ed (Optional[float], optional): span frequency end, in the unit Hz. The sweep frequency is (st, ed) plus measure frequency read from config. Defaults to None. Defaults to None.
        sweep_points (int, optional): sweep points. Defaults to 101.
        sideband (Optional[float], optional): sideband frequency. Defaults to None.

    Raises:
        ValueError: Args sideband is None here, which is illegal

    Returns:
        dict: config.
    """

    if sideband is None:
        raise ValueError('Args sideband is None here, which is illegal')
    sweep_list = generate_spanlist(
        center=center, delta=delta, st=st, ed=ed, sweep_points=sweep_points)
    measures = set(kernel.get(f'Q{i}.probe')for i in qubits)

    return {
        'init': {
            'name': 'S21',
            'qubits': qubits,
            'signal': 'raw',
            'scanner_name': 'S21_change_lo_without_constrains',
        },
        'setting': {
            'pre_setting': {},
            'compile_once': False,
            'circuit_type': 'cmds:qlispcmd',
            'circuit': [
                (('Measure', j), f'Q{i}') for j, i in enumerate(qubits)
            ],
        },
        'constrains': [],
        'sweep_config': dict(
            list({
                f'Q{i}': {
                    'addr': f'gate.Measure.Q{i}.params.frequency',
                } for i in qubits
            }.items())+list({
                m: {
                    'addr': f'{m}.setting.LO',
                } for m in measures
            }.items())
        ),
        'sweep_setting': {
            tuple([
                *[f'Q{i}' for i in qubits],
                *[m for m in measures],
            ]): [
                *[sweep_list for _ in qubits],
                *[sweep_list-sideband for _ in measures],
            ],
        },
        'plot_setting': [
            ['Frequency', np.asarray(
                [sweep_list+kernel.get(f'gate.Measure.Q{i}.params.frequency') for i in qubits]), 'Hz'],
        ],
        'plot': [
            ['Frequency', 'Hz'], ['Amplitude', 'a.u.'],
        ],
    }


def Spectrum_base_change_awg_without_constrains(qubits: list[int], delta: Optional[float] = None, st: Optional[float] = None, ed: Optional[float] = None,
                                                sweep_points: int = 101, pre_amp: Optional[float] = None, pre_duration: Optional[float] = None,
                                                pre_shape: Optional[str] = 'square', pre_alpha: Optional[float] = 1, pre_beta: Optional[float] = 0, pre_delta: Optional[float] = 0) -> dict:
    """
    [f'Q{i}'] Changing AWG waveform to measure spectrum.

    Args:
        qubits (list[int]): the qubit id.
        delta (Optional[float], optional): span frequency, in the unit Hz. The sweep frequency is (-delta, delta) plus measure frequency read from config. Defaults to None.
        st (Optional[float], optional): span frequency start, in the unit Hz. The sweep frequency is (st, ed) plus measure frequency read from config. Defaults to None.
        ed (Optional[float], optional): span frequency end, in the unit Hz. The sweep frequency is (st, ed) plus measure frequency read from config. Defaults to None. Defaults to None.
        sweep_points (int, optional): sweep points. Defaults to 101.
        pre_amp (Optional[float], optional): presetting drive amplitude, only valid in this measure. Defaults to None.
        pre_duration (Optional[float], optional): presetting drive duration, only valid in this measure. Defaults to None.
        pre_shape (Optional[str], optional): presetting drive shape, only valid in this measure. Defaults to 'square'.
        pre_alpha (Optional[float], optional): presetting drag alpha parameter, only valid in this measure. Defaults to 1.
        pre_beta (Optional[float], optional): presetting drag beta parameter, only valid in this measure. Defaults to 1.
        pre_delta (Optional[float], optional): presetting drag delta parameter, only valid in this measure. Defaults to 1.

    Returns:
        dict: config.
    """

    pre_setting_amp = {} if pre_amp is None else {
        f'gate.rfUnitary.Q{i}.params.amp': [[0, 0.5], [0, pre_amp]] for i in qubits
    }
    pre_setting_duration = {} if pre_duration is None else {
        f'gate.rfUnitary.Q{i}.params.duration': [[0, 0.5], [pre_duration, pre_duration]] for i in qubits
    }
    pre_setting_shape = {} if pre_shape is None else {
        f'gate.rfUnitary.Q{i}.params.shape': pre_shape for i in qubits
    }
    pre_setting_alpha = {} if pre_alpha is None else {
        f'gate.rfUnitary.Q{i}.params.alpha': pre_alpha for i in qubits
    }
    pre_setting_beta = {} if pre_beta is None else {
        f'gate.rfUnitary.Q{i}.params.beta': pre_beta for i in qubits
    }
    pre_setting_delta = {} if pre_delta is None else {
        f'gate.rfUnitary.Q{i}.params.delta': pre_delta for i in qubits
    }

    sweep_list = generate_spanlist(
        center=0, delta=delta, st=st, ed=ed, sweep_points=sweep_points)

    return {
        'init': {
            'name': 'Spectrum',
            'qubits': qubits,
            'signal': 'raw',
            'scanner_name': 'Spectrum_base_change_awg_without_constrains',
        },
        'setting': {
            'pre_setting': dict(
                list(pre_setting_amp.items())+list(
                    pre_setting_duration.items())+list(
                        pre_setting_shape.items())+list(pre_setting_alpha.items())+list(pre_setting_beta.items())+list(pre_setting_delta.items())),
            'compile_once': False,
            'circuit_type': 'cmds:qlispcmd',
            'circuit': [
                *[(('rfUnitary', np.pi, 0), f'Q{i}') for i in qubits],
                ('Barrier', tuple(f'Q{i}' for i in qubits)),
                *[(('Measure', j), f'Q{i}') for j, i in enumerate(qubits)],
            ],
        },
        'constrains': [],
        'sweep_config': {
            f'Q{i}': {
                'addr': f'gate.rfUnitary.Q{i}.params.frequency',
            } for i in qubits
        },
        'sweep_setting': {
            tuple([
                f'Q{i}' for i in qubits
            ]):  tuple([
                sweep_list+kernel.get(f'gate.rfUnitary.Q{i}.params.frequency') for i in qubits
            ]),
        },
        'plot_setting': [
            ['Frequency', np.asarray(
                [sweep_list+kernel.get(f'gate.rfUnitary.Q{i}.params.frequency') for i in qubits]), 'Hz'],
        ],
        'plot': [
            ['Frequency', 'Hz'], ['Amplitude', 'a.u.'],
        ],
    }


def Spectrum_base_change_lo_without_constrains(qubits: list[int], delta: Optional[float] = None, st: Optional[float] = None, ed: Optional[float] = None,
                                               sweep_points: int = 101, pre_amp: Optional[float] = None, pre_duration: Optional[float] = None,
                                               pre_shape: Optional[str] = 'square', pre_alpha: Optional[float] = 1, pre_beta: Optional[float] = 0, pre_delta: Optional[float] = 0) -> dict:
    """
    [f'Q{i}'] Changing lo frequency to measure spectrum.

    Args:
        qubits (list[int]): the qubit id.
        delta (Optional[float], optional): span frequency, in the unit Hz. The sweep frequency is (-delta, delta) plus measure frequency read from config. Defaults to None.
        st (Optional[float], optional): span frequency start, in the unit Hz. The sweep frequency is (st, ed) plus measure frequency read from config. Defaults to None.
        ed (Optional[float], optional): span frequency end, in the unit Hz. The sweep frequency is (st, ed) plus measure frequency read from config. Defaults to None. Defaults to None.
        sweep_points (int, optional): sweep points. Defaults to 101.
        pre_amp (Optional[float], optional): presetting drive amplitude, only valid in this measure. Defaults to None.
        pre_duration (Optional[float], optional): presetting drive duration, only valid in this measure. Defaults to None.
        pre_shape (Optional[str], optional): presetting drive shape, only valid in this measure. Defaults to 'square'.
        pre_alpha (Optional[float], optional): presetting drag alpha parameter, only valid in this measure. Defaults to 1.
        pre_beta (Optional[float], optional): presetting drag beta parameter, only valid in this measure. Defaults to 1.
        pre_delta (Optional[float], optional): presetting drag delta parameter, only valid in this measure. Defaults to 1.

    Returns:
        dict: config.
    """

    pre_setting_amp = {} if pre_amp is None else {
        f'gate.rfUnitary.Q{i}.params.amp': [[0, 0.5], [0, pre_amp]] for i in qubits
    }
    pre_setting_duration = {} if pre_duration is None else {
        f'gate.rfUnitary.Q{i}.params.duration': [[0, 0.5], [pre_duration, pre_duration]] for i in qubits
    }
    pre_setting_shape = {} if pre_shape is None else {
        f'gate.rfUnitary.Q{i}.params.shape': pre_shape for i in qubits
    }
    pre_setting_alpha = {} if pre_alpha is None else {
        f'gate.rfUnitary.Q{i}.params.alpha': pre_alpha for i in qubits
    }
    pre_setting_beta = {} if pre_beta is None else {
        f'gate.rfUnitary.Q{i}.params.beta': pre_beta for i in qubits
    }
    pre_setting_delta = {} if pre_delta is None else {
        f'gate.rfUnitary.Q{i}.params.delta': pre_delta for i in qubits
    }

    sweep_list = generate_spanlist(
        center=0, delta=delta, st=st, ed=ed, sweep_points=sweep_points)

    return {
        'init': {
            'name': 'Spectrum',
            'qubits': qubits,
            'signal': 'raw',
            'scanner_name': 'Spectrum_base_change_lo_without_constrains',
        },
        'setting': {
            'pre_setting': dict(
                list(pre_setting_amp.items())+list(
                    pre_setting_duration.items())+list(
                        pre_setting_shape.items())+list(pre_setting_alpha.items())+list(pre_setting_beta.items())+list(pre_setting_delta.items())),
            'compile_once': True,
            'circuit_type': 'cmds:qlispcmd',
            'circuit': [
                *[(('rfUnitary', np.pi, 0), f'Q{i}') for i in qubits],
                ('Barrier', tuple(f'Q{i}' for i in qubits)),
                *[(('Measure', j), f'Q{i}') for j, i in enumerate(qubits)],
            ],
        },
        'constrains': [],
        'sweep_config': dict(list({
            f'Q{i}': {
                'addr': f'gate.rfUnitary.Q{i}.params.frequency',
            } for i in qubits
        }.items())+list({
            f'LO{i}': {
                'addr': f'Q{i}.setting.LO',
            } for i in qubits
        }.items())),
        'sweep_setting': {
            tuple([
                *[f'Q{i}' for i in qubits],
                *[f'LO{i}' for i in qubits],
            ]): tuple([
                *[sweep_list+kernel.get(f'gate.rfUnitary.Q{i}.params.frequency') for i in qubits],
                *[sweep_list+kernel.get(f'Q{i}.setting.LO') for i in qubits],
            ]),
        },
        'plot_setting': [
            ['Frequency', np.asarray(
                [sweep_list+kernel.get(f'gate.rfUnitary.Q{i}.params.frequency') for i in qubits]), 'Hz'],
        ],
        'plot': [
            ['Frequency', 'Hz'], ['Amplitude', 'a.u.'],
        ],
    }


def Spectrum_base_change_lo_without_mixing_wave(qubits: list[int], delta: Optional[float] = None, st: Optional[float] = None, ed: Optional[float] = None,
                                                sweep_points: int = 101, pre_amp: Optional[float] = None, pre_duration: Optional[float] = None,
                                                pre_shape: Optional[str] = 'square', pre_alpha: Optional[float] = 1, pre_beta: Optional[float] = 0, pre_delta: Optional[float] = 0) -> dict:
    """
    [f'Q{i}'] Changing lo frequency to measure spectrum without any mixing frquency.

    Args:
        qubits (list[int]): the qubit id.
        delta (Optional[float], optional): span frequency, in the unit Hz. The sweep frequency is (-delta, delta) plus measure frequency read from config. Defaults to None.
        st (Optional[float], optional): span frequency start, in the unit Hz. The sweep frequency is (st, ed) plus measure frequency read from config. Defaults to None.
        ed (Optional[float], optional): span frequency end, in the unit Hz. The sweep frequency is (st, ed) plus measure frequency read from config. Defaults to None. Defaults to None.
        sweep_points (int, optional): sweep points. Defaults to 101.
        pre_amp (Optional[float], optional): presetting drive amplitude, only valid in this measure. Defaults to None.
        pre_duration (Optional[float], optional): presetting drive duration, only valid in this measure. Defaults to None.
        pre_shape (Optional[str], optional): presetting drive shape, only valid in this measure. Defaults to 'square'.
        pre_alpha (Optional[float], optional): presetting drag alpha parameter, only valid in this measure. Defaults to 1.
        pre_beta (Optional[float], optional): presetting drag beta parameter, only valid in this measure. Defaults to 1.
        pre_delta (Optional[float], optional): presetting drag delta parameter, only valid in this measure. Defaults to 1.

    Returns:
        dict: config.
    """

    pre_setting_amp = {} if pre_amp is None else {
        f'gate.rfUnitary.Q{i}.params.amp': [[0, 0.5], [0, pre_amp]] for i in qubits
    }
    pre_setting_duration = {} if pre_duration is None else {
        f'gate.rfUnitary.Q{i}.params.duration': [[0, 0.5], [pre_duration, pre_duration]] for i in qubits
    }
    pre_setting_shape = {} if pre_shape is None else {
        f'gate.rfUnitary.Q{i}.params.shape': pre_shape for i in qubits
    }
    pre_setting_alpha = {} if pre_alpha is None else {
        f'gate.rfUnitary.Q{i}.params.alpha': pre_alpha for i in qubits
    }
    pre_setting_beta = {} if pre_beta is None else {
        f'gate.rfUnitary.Q{i}.params.beta': pre_beta for i in qubits
    }
    pre_setting_delta = {} if pre_delta is None else {
        f'gate.rfUnitary.Q{i}.params.delta': pre_delta for i in qubits
    }

    sweep_list = generate_spanlist(
        center=0, delta=delta, st=st, ed=ed, sweep_points=sweep_points)

    lo_set = set()
    for i in qubits:
        lo = kernel.get(f'Q{i}.setting.LO')
        assert lo not in lo_set, 'Different qubits have the same local, which is not allowed.'
        lo_set.add(lo)

    return {
        'init': {
            'name': 'Spectrum',
            'qubits': qubits,
            'signal': 'raw',
            'scanner_name': 'Spectrum_base_change_lo_without_mixing_wave',
        },
        'setting': {
            'pre_setting': dict(
                list(pre_setting_amp.items())+list(
                    pre_setting_duration.items())+list(
                        pre_setting_shape.items())+list(pre_setting_alpha.items())+list(pre_setting_beta.items())+list(pre_setting_delta.items())),
            'compile_once': True,
            'circuit_type': 'cmds:qlispcmd',
            'circuit': [
                *[(('rfUnitary', np.pi, 0), f'Q{i}') for i in qubits],
                ('Barrier', tuple(f'Q{i}' for i in qubits)),
                *[(('Measure', j), f'Q{i}') for j, i in enumerate(qubits)],
            ],
        },
        'constrains': [],
        'sweep_config': dict(list({
            f'Q{i}': {
                'addr': f'gate.rfUnitary.Q{i}.params.frequency',
            } for i in qubits
        }.items())+list({
            f'LO{i}': {
                'addr': f'Q{i}.setting.LO',
            } for i in qubits
        }.items())),
        'sweep_setting': {
            tuple([
                *[f'Q{i}' for i in qubits],
                *[f'LO{i}' for i in qubits],
            ]): [
                *[sweep_list+kernel.get(f'gate.rfUnitary.Q{i}.params.frequency') for i in qubits],
                *[sweep_list+kernel.get(f'gate.rfUnitary.Q{i}.params.frequency') for i in qubits],
            ],
        },
        'plot_setting': [
            ['Frequency', np.asarray(
                [sweep_list+kernel.get(f'gate.rfUnitary.Q{i}.params.frequency') for i in qubits]), 'Hz'],
        ],
        'plot': [
            ['Frequency', 'Hz'], ['Amplitude', 'a.u.'],
        ],
    }


def PowerRabi_base_scaleTimes(qubits: list[int], sweep_points: int = 21, scale_times: float = 4, pre_duration: Optional[float] = None,
                              signal: str = 'raw') -> dict:
    """
    [f'Q{i}'] power Rabi.

    Args:
        qubits (list[int]): the qubit id.
        sweep_points (int, optional): sweep points. Defaults to 21.
        scale_times (float, optional): raion of max scale in this measure to the original. Defaults to 4.
        pre_duration (Optional[float], optional): presetting drive duration. Defaults to None.
        signal (str, optional): signal type. Defaults to 'raw'.

    Returns:
        dict: config.
    """

    pre_setting_duration = {} if pre_duration is None else {
        f'gate.rfUnitary.Q{i}.params.duration': [[0, 1], [pre_duration, pre_duration]] for i in qubits
    }

    return {
        'init': {
            'name': 'PowerRabi',
            'qubits': qubits,
            'signal': signal,
            'scanner_name': 'PowerRabi_base_scaleTimes',
        },
        'setting': {
            'pre_setting': pre_setting_duration,
            'compile_once': False,
            'circuit_type': 'cmds:qlispcmd',
            'circuit': [
                *[(('rfUnitary', np.pi, 0), f'Q{i}') for i in qubits],
                ('Barrier', tuple(f'Q{i}' for i in qubits)),
                *[(('Measure', j), f'Q{i}') for j, i in enumerate(qubits)],
            ],
        },
        'constrains': [],
        'sweep_config': dict(list({
            f'Q{i}_amp': {
                'addr': f'gate.rfUnitary.Q{i}.params.amp',
            } for i in qubits
        }.items())+list({
            f'Q{i}': {
                'addr': None,
            } for i in qubits
        }.items())),
        'sweep_setting': {
            tuple([
                f'Q{i}' for i in qubits
            ]): tuple([
                np.linspace(0, min(1, scale_times*kernel.get(f'gate.rfUnitary.Q{i}.params.amp')[1][-1]), sweep_points) for i in qubits
            ]),
        },
        'sweep_addition': {
            f'Q{i}_amp': lambda i=i, **kw: [[0, 0.5], [0, kw[f'Q{i}']]] for i in qubits
        },
        'plot_setting': [
            ['Scale', np.asarray([np.linspace(0, min(1, scale_times*kernel.get(f'gate.rfUnitary.Q{i}.params.amp')[1][-1]), sweep_points) for i in qubits]),
             'a.u.'],
        ],
        'plot': [
            ['Scale', 'a.u.'], ['Probability' if signal ==
                                'state' else 'Amplitude', 'a.u.'],
        ],
    }


def PowerRabi_base_n_pulse(qubits: list[int], sweep_points: int = 21, n: int = 1, signal: str = 'raw', mode: str = 'log') -> dict:
    """
    [f'Q{i}'] n pulses of power Rabi.

    Args:
        qubits (list[int]): the qubit id.
        sweep_points (int, optional): sweep points. Defaults to 21.
        n (int, optional): numbers of pulses. Defaults to 1.
        mode (str, optional): in ['linear' (for linear time sampling), 'log' (for log time sampling)]. Defaults to 'linear'.

    Returns:
        dict: config.
    """

    sweep_list = []
    if mode=='linear':
        for i in qubits:
            scale = kernel.get(f'gate.rfUnitary.Q{i}.params.amp')[-1][-1]
            sweep_list.append(np.linspace((1-1/n)*scale, min(1, (1+1/n)*scale), sweep_points))
    elif mode=='log':
        for i in qubits:
            scale = kernel.get(f'gate.rfUnitary.Q{i}.params.amp')[-1][-1]
            sweep_list.append(generate_spanlist_log(center=scale, st=scale*(1-1/n), ed=min(1, scale*(1+1/n)), sweep_points=sweep_points))
    else:
        raise ValueError(f'Mode {mode} is not supported.')

    return {
        'init': {
            'name': 'PowerRabi',
            'qubits': qubits,
            'signal': signal,
            'scanner_name': 'PowerRabi_base_n_pulse'
        },
        'setting': {
            'pre_setting': {},
            'compile_once': False,
            'circuit_type': 'cmds:qlispcmd',
            'circuit': [
                *[(('rfUnitary', np.pi, 0), f'Q{i}') for i in qubits]*n,
                ('Barrier', tuple(f'Q{i}' for i in qubits)),
                *[(('Measure', j), f'Q{i}') for j, i in enumerate(qubits)],
            ],
        },
        'constrains': [],
        'sweep_config': dict(list({
            f'Q{i}_amp': {
                'addr': f'gate.rfUnitary.Q{i}.params.amp',
            } for i in qubits
        }.items())+list({
            f'Q{i}': {
                'addr': None,
            } for i in qubits
        }.items())),
        'sweep_setting': {
            tuple([
                f'Q{i}' for i in qubits
            ]): tuple([
                sweep_list[j] for j, i in enumerate(qubits)
            ]),
        },
        'sweep_addition': {
            f'Q{i}_amp': lambda i=i, **kw: [[0, 0.5], [0.00001, kw[f'Q{i}']]] for i in qubits
        },
        'plot_setting': [
            ['Scale', np.asarray(sweep_list), 'a.u.'],
        ],
        'plot': [
            ['Scale', 'a.u.'], ['Probability' if signal ==
                                'state' else 'Amplitude', 'a.u.'],
        ],
    }


def TimeRabi_base_durationTimes(qubits: list[int], sweep_points: int = 51, duration_times: float = 4, pre_amp: Optional[float] = None,
                                signal: str = 'raw') -> dict:
    """
    [f'Q{i}'] time Rabi.

    Args:
        qubits (list[int]): the qubit id.
        sweep_points (int, optional): sweep points. Defaults to 51.
        scale_times (float, optional): raion of max scale in this measure to the original. Defaults to 4.
        pre_duration (Optional[float], optional): presetting drive duration. Defaults to None.
        signal (str, optional): signal type. Defaults to 'raw'.

    Returns:
        dict: config.
    """

    pre_setting_amp = {} if pre_amp is None else {
        f'gate.rfUnitary.Q{i}.params.amp': [[0, 1], [0, pre_amp]] for i in qubits
    }

    return {
        'init': {
            'name': 'TimeRabi',
            'qubits': qubits,
            'signal': signal,
            'scanner_name': 'TimeRabi_base_durationTimes',
        },
        'setting': {
            'pre_setting': pre_setting_amp,
            'compile_once': False,
            'circuit_type': 'cmds:qlispcmd',
            'circuit': [
                *[(('Delay', 20e-6), f'Q{i}') for i in qubits],
                *[(('rfUnitary', np.pi, 0), f'Q{i}') for i in qubits],
                ('Barrier', tuple(f'Q{i}' for i in qubits)),
                *[(('Measure', j), f'Q{i}') for j, i in enumerate(qubits)],
            ],
        },
        'constrains': [],
        'sweep_config': dict(list({
            f'Q{i}_duration': {
                'addr': f'gate.rfUnitary.Q{i}.params.duration',
            } for i in qubits
        }.items())+list({
            f'Q{i}': {
                'addr': None,
            } for i in qubits
        }.items())),
        'sweep_setting': {
            tuple([
                f'Q{i}' for i in qubits
            ]): tuple([
                np.linspace(2e-9, 2e-9+duration_times*kernel.get(f'gate.rfUnitary.Q{i}.params.duration')[1][-1], sweep_points) for i in qubits
            ]),
        },
        'sweep_addition': {
            f'Q{i}_duration': lambda i=i, **kw: [[0, 1], [kw[f'Q{i}'], kw[f'Q{i}']]] for i in qubits
        },
        'plot_setting': [
            ['Time', np.asarray([np.linspace(0,
                                             duration_times*kernel.get(f'gate.rfUnitary.Q{i}.params.duration')[1][-1], sweep_points) for i in qubits]), 's'],
        ],
        'plot': [
            ['Time', 's'], ['Probability' if signal ==
                            'state' else 'Amplitude', 'a.u.'],
        ],
    }


def Ramsey_without_constrains(qubits: list[int], rotate: float, time_length: float, points_in_one_period: int = 10, signal: str = 'raw') -> dict:
    """
    ['time'] Ramsey.

    Args:
        qubits (list[int]): the qubit id.
        rotate (float): offset of frequency.
        time_length (float): length of delay time.
        points_in_one_period (int, optional): points per period. Defaults to 10.
        signal (str, optional): signal type. Defaults to 'raw'.

    Returns:
        dict: config.
    """

    from waveforms.quantum import Ramsey

    return {
        'init': {
            'name': 'Ramsey',
            'qubits': qubits,
            'signal': signal,
            'scanner_name': 'Ramsey_without_constrains',
        },
        'setting': {
            'pre_setting': {},
            'compile_once': False,
            'circuit_type': 'cmds:qlispcmd',
            'circuit': [],
        },
        'constrains': [],
        'sweep_config': {
            'time': {
                'addr': None,
            },
            'circuit': {
                'addr': 'setting.circuit',
            },
        },
        'sweep_setting': {
            'time': np.linspace(0, time_length, round(time_length*rotate*points_in_one_period)+1),
        },
        'sweep_addition': {
            'circuit': lambda time: [
                *[(('Delay', 20e-6), f'Q{i}') for i in qubits],
                *chain.from_iterable(Ramsey(qubit=f'Q{i}', t=time, f=rotate) for i in qubits),
                ('Barrier', tuple(f'Q{i}' for i in qubits)),
                *[(('Measure', j), f'Q{i}') for j, i in enumerate(qubits)],
            ],
        },
        'plot_setting': [
            ['Time', np.asarray([np.linspace(0, time_length, round(
                time_length*rotate*points_in_one_period)+1) for _ in qubits]), 's'],
        ],
        'plot': [
            ['Delay', 's'], ['Probability' if signal ==
                             'state' else 'Amplitude', 'a.u.'],
        ],
    }


def T1_without_constrains(qubits: list[int], time_length: Union[float, list], sweep_points: int = 51, mode: str = 'linear', signal: str = 'raw') -> dict:
    """
    ['f'Q{i}'] T1 Measure.

    Args:
        qubits (list[int]): the qubit id.
        time_length (Union[float, list]): the delay time length, for each qubit(in the type list of float) or for all qubits(int the type float).
        sweep_points (int, optional): sweep points. Defaults to 51.
        mode (str, optional): in ['linear' (for linear time sampling), 'log' (for log time sampling)]. Defaults to 'linear'.
        step (float, optional): the minimum step size in log time sampling. Defaults to 16e-9.
        signal (str, optional): signal type. Defaults to 'raw'.

    Raises:
        ValueError: The time_length is illegal.
        ValueError: [description]

    Returns:
        dict: config.
    """

    if isinstance(time_length, list) and len(time_length) != len(qubits):
        raise ValueError('The time_length is illegal')
    elif isinstance(time_length, float):
        time_length = [time_length for _ in qubits]
    else:
        raise ValueError('The time_length is illegal')

    timeList = []
    if mode in ['log']:
        for j, i in enumerate(qubits):
            sr = kernel.get(f'Q{i}.waveform.SR')
            timeList.append(np.logspace(
                0, np.log10(time_length*sr), sweep_points)/sr)
    else:
        for j, i in enumerate(qubits):
            timeList.append(np.concatenate((np.linspace(0, time_length[j]/3, sweep_points//3*2),
                            np.linspace(time_length[j]/3, time_length[j], sweep_points-sweep_points//3*2+1)[1:])))

    return {
        'init': {
            'name': 'T1',
            'qubits': qubits,
            'signal': signal,
            'scanner_name': 'T1_without_constrains',
        },
        'setting': {
            'pre_setting': {},
            'compile_once': False,
            'circuit_type': 'cmds:qlispcmd',
            'circuit': [],
        },
        'constrains': [],
        'sweep_config': dict(list({
            f'Q{i}': {
                'addr': None,
            } for i in qubits
        }.items())+list({
            'circuit': {
                'addr': f'setting.circuit',
            },
        }.items())),
        'sweep_setting': {
            tuple([
                f'Q{i}' for i in qubits
            ]): tuple([
                timeList[j] for j, i in enumerate(qubits)
            ]),
        },
        'sweep_addition': {
            'circuit': lambda **kw: [
                *[(('rfUnitary', np.pi, 0), f'Q{i}') for i in qubits],
                *[(('Delay', kw[f'Q{i}']), f'Q{i}')
                  for j, i in enumerate(qubits)],
                ('Barrier', tuple(f'Q{i}' for i in qubits)),
                *[(('Measure', j), f'Q{i}') for j, i in enumerate(qubits)],
            ],
        },
        'plot_setting': [
            ['Time', np.asarray(timeList), 's'],
        ],
        'plot': [
            ['Delay', 's'], ['Probability' if signal ==
                             'state' else 'Amplitude', 'a.u.'],
        ],
    }


def SpinEcho_without_constrains(qubits: list[int], rotate: float, time_length: float, points_in_one_period: int = 10, signal: str = 'raw') -> dict:
    """
    ['time'] Spin echo.

    Args:
        qubits (list[int]): the qubit id.
        rotate (float): offset of frequency.
        time_length (float): length of delay time.
        points_in_one_period (int, optional): points per period. Defaults to 10.
        signal (str, optional): signal type. Defaults to 'raw'.

    Returns:
        dict: config.
    """

    from waveforms.quantum import SpinEcho

    return {
        'init': {
            'name': 'SpinEcho',
            'qubits': qubits,
            'signal': signal,
            'scanner_name': 'SpinEcho_without_constrains',
        },
        'setting': {
            'pre_setting': {},
            'compile_once': False,
            'circuit_type': 'cmds:qlispcmd',
            'circuit': [],
        },
        'constrains': [],
        'sweep_config': {
            'time': {
                'addr': None,
            },
            'circuit': {
                'addr': 'setting.circuit',
            },
        },
        'sweep_setting': {
            'time': np.linspace(0, time_length, round(time_length*rotate*points_in_one_period)+1),
        },
        'sweep_addition': {
            'circuit': lambda time: [
                *chain.from_iterable(SpinEcho(qubit=f'Q{i}', t=time, f=rotate) for i in qubits),
                ('Barrier', tuple(f'Q{i}' for i in qubits)),
                *[(('Measure', j), f'Q{i}') for j, i in enumerate(qubits)],
            ],
        },
        'plot_setting': [
            ['Time', np.asarray([np.linspace(0, time_length, round(
                time_length*rotate*points_in_one_period)+1) for _ in qubits]), 's'],
        ],
        'plot': [
            ['Delay', 's'], ['Probability' if signal ==
                             'state' else 'Amplitude', 'a.u.'],
        ],
    }


def CPMG_without_constrains(qubits: list[int], rotate: float, time_length: float, n: int, points_in_one_period: int = 10, signal: str = 'raw') -> dict:
    """
    ['time'] CPMG.

    Args:
        qubits (list[int]): the qubit id.
        rotate (float): offset of frequency.
        time_length (float): length of delay time.
        points_in_one_period (int, optional): points per period. Defaults to 10.
        signal (str, optional): signal type. Defaults to 'raw'.

    Returns:
        dict: config.
    """

    from waveforms.quantum import CPMG

    return {
        'init': {
            'name': 'CPMG',
            'qubits': qubits,
            'signal': signal,
            'scanner_name': 'CPMG_without_constrains',
        },
        'setting': {
            'pre_setting': {},
            'compile_once': False,
            'circuit_type': 'cmds:qlispcmd',
            'circuit': [],
        },
        'constrains': [],
        'sweep_config': {
            'time': {
                'addr': None,
            },
            'circuit': {
                'addr': 'setting.circuit',
            },
        },
        'sweep_setting': {
            'time': np.linspace(0, time_length, round(time_length*rotate*points_in_one_period)+1),
        },
        'sweep_addition': {
            'circuit': lambda time: [
                *chain.from_iterable(CPMG(qubit=f'Q{i}', t=time, f=rotate, n=n) for i in qubits),
                ('Barrier', tuple(f'Q{i}' for i in qubits)),
                *[(('Measure', j), f'Q{i}') for j, i in enumerate(qubits)],
            ],
        },
        'plot_setting': [
            ['Time', np.asarray([np.linspace(0, time_length, round(
                time_length*rotate*points_in_one_period)+1) for _ in qubits]), 's'],
        ],
        'plot': [
            ['Delay', 's'], ['Probability' if signal ==
                             'state' else 'Amplitude', 'a.u.'],
        ],
    }


def CP_without_constrains(qubits: list[int], rotate: float, time_length: float, n: int, points_in_one_period: int = 10, signal: str = 'raw') -> dict:
    """
    ['time'] CP.

    Args:
        qubits (list[int]): the qubit id.
        rotate (float): offset of frequency.
        time_length (float): length of delay time.
        points_in_one_period (int, optional): points per period. Defaults to 10.
        signal (str, optional): signal type. Defaults to 'raw'.

    Returns:
        dict: config.
    """

    from waveforms.quantum import CP

    return {
        'init': {
            'name': 'CP',
            'qubits': qubits,
            'signal': signal,
            'scanner_name': 'CP_without_constrains',
        },
        'setting': {
            'pre_setting': {},
            'compile_once': False,
            'circuit_type': 'cmds:qlispcmd',
            'circuit': [],
        },
        'constrains': [],
        'sweep_config': {
            'time': {
                'addr': None,
            },
            'circuit': {
                'addr': 'setting.circuit',
            },
        },
        'sweep_setting': {
            'time': np.linspace(0, time_length, round(time_length*rotate*points_in_one_period)+1),
        },
        'sweep_addition': {
            'circuit': lambda time: [
                *chain.from_iterable(CP(qubit=f'Q{i}', t=time, f=rotate, n=n) for i in qubits),
                ('Barrier', tuple(f'Q{i}' for i in qubits)),
                *[(('Measure', j), f'Q{i}') for j, i in enumerate(qubits)],
            ],
        },
        'plot_setting': [
            ['Time', np.asarray([np.linspace(0, time_length, round(
                time_length*rotate*points_in_one_period)+1) for _ in qubits]), 's'],
        ],
        'plot': [
            ['Delay', 's'], ['Probability' if signal ==
                             'state' else 'Amplitude', 'a.u.'],
        ],
    }


def AllXY_without_constrains(qubits: list[int], signal: str = 'raw', repeat: int = 1, n: int = 0) -> dict:
    """
    [21] AllXY 21 circuit

    Args:
        qubits (list[int]): the qubit id.
        signal (str, optional): signal type. Defaults to 'raw'.
        repeat (int, optional): repeat times. Defaults to 1.

    Returns:
        dict: config.
    """

    from waveforms.quantum import ALLXY

    def span_n(x, n, changeI: float = None):
        ret = []
        for item in x:
            n_pre = 4*n+1
            if item[0] == 'I' and changeI is not None:
                item = list(item)
                item[0] = ('Delay', changeI)
                item = tuple(item)
            if item[0] in ['X']:
                item = list(item)
                item[0] = ('rfUnitary', np.pi, 0)
                item = tuple(item)
                n_pre = 2*n+1
            if item[0] in ['Y']:
                item = list(item)
                item[0] = ('rfUnitary', np.pi, np.pi/2)
                item = tuple(item)
                n_pre = 2*n+1
            if item[0] in ['X/2']:
                item = list(item)
                item[0] = ('rfUnitary', np.pi/2, 0)
                item = tuple(item)
            if item[0] in ['Y/2']:
                item = list(item)
                item[0] = ('rfUnitary', np.pi/2, np.pi/2)
                item = tuple(item)
            ret.extend([item]*n_pre)
        return ret

    return {
        'init': {
            'name': 'AllXY',
            'qubits': qubits,
            'signal': signal,
            'scanner_name': 'AllXY_without_constrains',
            'shots': round(repeat)*1024,
        },
        'setting': {
            'pre_setting': {},
            'compile_once': False,
            'circuit_type': 'cmds:qlispcmd',
            'circuit': [],
        },
        'constrains': [],
        'sweep_config': {
            'circuit': {
                'addr': 'setting.circuit',
            },
            'number': {
                'addr': None,
            }
        },
        'sweep_setting': {
            'number': list(range(21)),
        },
        'sweep_addition': {
            'circuit': lambda number: [
                *chain.from_iterable(span_n(ALLXY(qubit=f'Q{i}', i=number), n,
                                            kernel.get(f'gate.rfUnitary.Q{i}.params.duration')[-1][-1]+kernel.get(f'gate.rfUnitary.Q{i}.params.buffer')) for i in qubits),
                ('Barrier', tuple(f'Q{i}' for i in qubits)),
                *[(('Measure', j), f'Q{i}') for j, i in enumerate(qubits)],
            ],
        },
        'plot': [
            ['Item', 'a.u.'], ['Probability' if signal ==
                               'state' else 'Amplitude', 'a.u.'],
        ],
    }


def ReadoutFrequency_without_constrains(qubits: list[int], delta: Optional[float] = None, st: Optional[float] = None, ed: Optional[float] = None,
                                        sweep_points: int = 101) -> dict:
    """
    [f'Q{i}] Read frequency measurement under 0 state and 1 state.

    Args:
        qubits (list[int]): the qubit id.
        delta (Optional[float], optional): span frequency, in the unit Hz. The sweep frequency is (-delta, delta) plus measure frequency read from config. Defaults to None.
        st (Optional[float], optional): span frequency start, in the unit Hz. The sweep frequency is (st, ed) plus measure frequency read from config. Defaults to None.
        ed (Optional[float], optional): span frequency end, in the unit Hz. The sweep frequency is (st, ed) plus measure frequency read from config. Defaults to None. Defaults to None.
        sweep_points (int, optional): sweep points. Defaults to 101.

    Returns:
        dict: config.
    """

    sweep_list = generate_spanlist(
        center=0, delta=delta, st=st, ed=ed, sweep_points=sweep_points)
    measures = set(kernel.get(f'Q{i}.probe')for i in qubits)

    return {
        'init': {
            'name': 'ReadoutFrequency',
            'qubits': qubits,
            'signal': 'raw',
            'scanner_name': 'ReadoutFrequency_without_constrains',
        },
        'setting': {
            'pre_setting': {},
            'compile_once': False,
            'circuit_type': 'cmds:qlispcmd',
            'circuit': [],
        },
        'constrains': [],
        'sweep_config': dict(
            list({
                f'Q{i}': {
                    'addr': f'gate.Measure.Q{i}.params.frequency',
                } for i in qubits
            }.items())+list({
                m: {
                    'addr': f'{m}.setting.LO',
                } for m in measures
            }.items())+list({
                'skip': {
                    'addr': f'setting.skip'
                },
                'circuit': {
                    'addr': f'setting.circuit'
                }
            }.items())
        ),
        'sweep_setting': {
            ('circuit', 'skip'): [[[
                (('Measure', j), f'Q{i}') for j, i in enumerate(qubits)
            ], [
                *[(('rfUnitary', np.pi, 0), f'Q{i}') for i in qubits],
                ('Barrier', tuple(f'Q{i}' for i in qubits)),
                *[(('Measure', j), f'Q{i}') for j, i in enumerate(qubits)]
            ], ], [1, 2]],
            tuple([
                *[f'Q{i}' for i in qubits],
                *[m for m in measures],
            ]): [
                *[sweep_list+kernel.get(f'gate.Measure.Q{i}.params.frequency') for i in qubits],
                *[sweep_list+kernel.get(f'{m}.setting.LO') for m in measures],
            ],
        },
        'plot_setting': [
            ['State', np.array([[0, 1] for _ in qubits]), 's'],
            ['Frequency', np.array(
                [sweep_list+kernel.get(f'gate.Measure.Q{i}.params.frequency') for i in qubits]), 'Hz'],
        ],
        'plot': [
            ['Frequency', 'Hz'], ['Amplitude', 'a.u.'],
        ],
    }


def ReadoutAmp_without_constrains(qubits: list[int], st: float, ed: float, sweep_points: int = 101) -> dict:
    """
    [f'Q{i}] Read amp measurement under 0 state and 1 state. 

    Args:
        qubits (list[int]): the qubit id.
        st (float, optional): span lower times bound. The sweep amp is (st, ed)*amp plus measure amp read from config.
        ed (float, optional): span upper times bound. The sweep amp is (st, ed)*amp plus measure amp read from config.
        sweep_points (int, optional): sweep points. Defaults to 101.

    Returns:
        dict: config.
    """

    return {
        'init': {
            'name': 'ReadoutAmp',
            'qubits': qubits,
            'signal': 'raw',
            'scanner_name': 'ReadoutAmp_without_constrains',
        },
        'setting': {
            'pre_setting': {},
            'compile_once': False,
            'circuit_type': 'cmds:qlispcmd',
            'circuit': [],
        },
        'constrains': [],
        'sweep_config': dict(
            list({
                f'Q{i}': {
                    'addr': f'gate.Measure.Q{i}.params.amp',
                } for i in qubits
            }.items())+list({
                'circuit': {
                    'addr': f'setting.circuit'
                }
            }.items())
        ),
        'sweep_setting': {
            'circuit': [[
                (('Measure', j), f'Q{i}') for j, i in enumerate(qubits)
            ], [
                *[(('rfUnitary', np.pi, 0), f'Q{i}') for i in qubits],
                ('Barrier', tuple(f'Q{i}' for i in qubits)),
                *[(('Measure', j), f'Q{i}') for j, i in enumerate(qubits)]
            ], ],
            tuple([
                f'Q{i}' for i in qubits
            ]): [
                np.linspace(st, ed, sweep_points)*kernel.get(f'gate.Measure.Q{i}.params.amp') for i in qubits
            ],
        },
        'plot': [
            ['Scale', 'a.u.'], ['Amplitude', 'a.u.'],
        ],
    }


def Scatter2(qubits: list[int], repeat: int = 1) -> dict:
    """
    [2] 0 state and 1 state classification.

    Args:
        qubits (list[int]): the qubit id.
        repeat (int, optional): repeat times. Defaults to 1.

    Returns:
        dict: config.
    """

    return {
        'init': {
            'name': 'Scatter2',
            'qubits': qubits,
            'signal': 'raw',
            'scanner_name': 'Scatter2',
            'shots': round(repeat)*1024
        },
        'setting': {
            'pre_setting': {},
            'compile_once': False,
            'circuit_type': 'cmds:qlispcmd',
            'circuit': [],
        },
        'constrains': [],
        'sweep_config': {
            'circuit': {
                'addr': f'setting.circuit',
            },
        },
        'sweep_setting': {
            'circuit': [[
                *[(('Delay', 5e-6), f'Q{i}') for i in qubits],
                *[('X/2', f'Q{i}') for i in qubits],
                *[('-X/2', f'Q{i}') for i in qubits],
                ('Barrier', tuple(f'Q{i}' for i in qubits)),
                *[(('Measure', j), f'Q{i}') for j, i in enumerate(qubits)]
            ], [
                *[(('Delay', 5e-6), f'Q{i}') for i in qubits],
                *[('X/2', f'Q{i}') for i in qubits],
                *[('X/2', f'Q{i}') for i in qubits],
                ('Barrier', tuple(f'Q{i}' for i in qubits)),
                *[(('Measure', j), f'Q{i}') for j, i in enumerate(qubits)]
            ], ],
        },
    }


def AllXY6and16_alpha_without_constrains(qubits: list[int], delta: Optional[float] = None, st: Optional[float] = None, ed: Optional[float] = None,
                                         sweep_points: int = 41, repeat: int = 1, signal: str = 'raw') -> dict:
    """
    [f'Q{i}'] change drag.alpha and measure AllXY 6 and AllXY 16 to determin the `rfUnitary.params.alpha`

    Args:
        qubits (list[int]): the qubit id.
        delta (Optional[float], optional): span alpha coefficient, in the dimension 1. Defaults to None.
        st (Optional[float], optional): start alpha coefficient, in the dimension 1. Defaults to None.
        ed (Optional[float], optional): end alpha coefficient, in the dimension 1. Defaults to None.
        sweep_points (int, optional): sweep points. Defaults to 41.
        repeat (int, optional): repeat times. Defaults to 1.
        signal (str, optional): signal type. Defaults to 'raw'.

    Returns:
        dict: config.
    """

    sweep_list = generate_spanlist(
        center=1, delta=delta, st=st, ed=ed, sweep_points=2)
    base = [kernel.get(f'gate.rfUnitary.Q{i}.params.alpha') for i in qubits]

    return {
        'init': {
            'name': 'AllXY6and16_alpha',
            'qubits': qubits,
            'signal': signal,
            'scanner_name': 'AllXY6and16_alpha_without_constrains',
            'shots': round(repeat)*1024,
        },
        'setting': {
            'pre_setting': {},
            'compile_once': False,
            'circuit_type': 'cmds:qlispcmd',
            'circuit': [],
        },
        'constrains': [],
        'sweep_config': dict(
            list({
                f'Q{i}': {
                    'addr': f'gate.rfUnitary.Q{i}.params.alpha',
                } for i in qubits
            }.items())+list({
                'circuit': {
                    'addr': f'setting.circuit'
                }
            }.items())
        ),
        'sweep_setting': {
            'circuit': [[[
                *[('Y/2', f'Q{i}') for i in qubits],
                *[(('Delay', kernel.get(f'gate.rfUnitary.Q{i}.params.duration')[1][0]), f'Q{i}') for i in qubits],
                ('Barrier', tuple(f'Q{i}' for i in qubits)),
                *[(('Measure', j), f'Q{i}') for j, i in enumerate(qubits)],
            ], [
                *[('Y', f'Q{i}') for i in qubits],
                *[('Y/2', f'Q{i}') for i in qubits],
                ('Barrier', tuple(f'Q{i}' for i in qubits)),
                *[(('Measure', j), f'Q{i}') for j, i in enumerate(qubits)],
            ], ], ],
            tuple([
                *[f'Q{i}' for i in qubits],
            ]): [np.linspace(
                max(0, sweep_list[0]*base[j]), min(1,
                                                   sweep_list[1]*base[j]), sweep_points
            ) for j, i in enumerate(qubits)],
        },
        'plot': [
            ['Scale', 'a.u.'], ['Probability' if signal ==
                                'state' else 'Amplitude', 'a.u.'],
        ],
    }


def AllXY7and8_delta_without_constrains(qubits: list[int], delta: Optional[float] = None, st: Optional[float] = None, ed: Optional[float] = None,
                                        sweep_points: int = 41, repeat: int = 1, signal: str = 'raw') -> dict:
    """
    [f'Q{i}'] change drag.delta and measure AllXY 7 and AllXY 8 to determin the `rfUnitary.params.delta`

    Args:
        qubits (list[int]): the qubit id.
        delta (Optional[float], optional): span delta coefficient, in the dimension frequency, Hz. Defaults to None.
        st (Optional[float], optional): start delta coefficient, in the dimension frequency, Hz. Defaults to None.
        ed (Optional[float], optional): end delta coefficient, in the dimension frequency, Hz. Defaults to None.
        sweep_points (int, optional): sweep points. Defaults to 41.
        repeat (int, optional): repeat times. Defaults to 1.
        signal (str, optional): signal type. Defaults to 'raw'.

    Returns:
        dict: [description]
    """

    sweep_list = generate_spanlist(
        center=0, delta=delta, st=st, ed=ed, sweep_points=sweep_points)

    return {
        'init': {
            'name': 'AllXY7and8_delta',
            'qubits': qubits,
            'signal': signal,
            'scanner_name': 'AllXY7and8_delta_without_constrains',
            'shots': round(repeat)*1024,
        },
        'setting': {
            'pre_setting': {},
            'compile_once': False,
            'circuit_type': 'cmds:qlispcmd',
            'circuit': [],
        },
        'constrains': [],
        'sweep_config': dict(
            list({
                f'Q{i}': {
                    'addr': f'gate.rfUnitary.Q{i}.params.delta',
                } for i in qubits
            }.items())+list({
                'circuit': {
                    'addr': f'setting.circuit'
                }
            }.items())
        ),
        'sweep_setting': {
            'circuit': [[
                *[('X/2', f'Q{i}') for i in qubits],
                *[('Y/2', f'Q{i}') for i in qubits],
                ('Barrier', tuple(f'Q{i}' for i in qubits)),
                *[(('Measure', j), f'Q{i}') for j, i in enumerate(qubits)]
            ], [
                *[('Y/2', f'Q{i}') for i in qubits],
                *[('X/2', f'Q{i}') for i in qubits],
                ('Barrier', tuple(f'Q{i}' for i in qubits)),
                *[(('Measure', j), f'Q{i}') for j, i in enumerate(qubits)]
            ], ],
            tuple([
                f'Q{i}' for i in qubits
            ]): [
                sweep_list+kernel.get(f'gate.rfUnitary.Q{i}.params.delta') for i in qubits
            ],
        },
        'plot': [
            ['Delta', 'Hz'], ['Probability' if signal ==
                              'state' else 'Amplitude', 'a.u.'],
        ],
    }


def AllXY11and12_beta_without_constrains(qubits: list[int], delta: Optional[float] = None, st: Optional[float] = None, ed: Optional[float] = None,
                                         sweep_points: int = 41, repeat: int = 1, signal: str = 'raw') -> dict:
    """
    [f'Q{i}'] change drag.beta and measure AllXY 11 and AllXY 12 to determin the `rfUnitary.params.beta`

    Args:
        qubits (list[int]): the qubit id.
        delta (Optional[float], optional): span beta coefficient, in the dimension 1e-9. Defaults to None.
        st (Optional[float], optional): start beta coefficient, in the dimension 1e-9. Defaults to None.
        ed (Optional[float], optional): end beta coefficient, in the dimension 1e-9. Defaults to None.
        sweep_points (int, optional): sweep points. Defaults to 41.
        repeat (int, optional): repeat times. Defaults to 1.
        signal (str, optional): signal type. Defaults to 'raw'.

    Returns:
        dict: [description]
    """

    sweep_list = generate_spanlist(
        center=0, delta=delta, st=st, ed=ed, sweep_points=sweep_points)

    return {
        'init': {
            'name': 'AllXY11and12_beta',
            'qubits': qubits,
            'signal': signal,
            'scanner_name': 'AllXY11and12_beta_without_constrains',
            'shots': round(repeat)*1024,
        },
        'setting': {
            'pre_setting': {},
            'compile_once': False,
            'circuit_type': 'cmds:qlispcmd',
            'circuit': [],
        },
        'constrains': [],
        'sweep_config': dict(
            list({
                f'Q{i}': {
                    'addr': f'gate.rfUnitary.Q{i}.params.beta',
                } for i in qubits
            }.items())+list({
                'circuit': {
                    'addr': f'setting.circuit'
                }
            }.items())
        ),
        'sweep_setting': {
            'circuit': [[
                *[('X', f'Q{i}') for i in qubits],
                *[('Y/2', f'Q{i}') for i in qubits],
                ('Barrier', tuple(f'Q{i}' for i in qubits)),
                *[(('Measure', j), f'Q{i}') for j, i in enumerate(qubits)],
            ], [
                *[('Y', f'Q{i}') for i in qubits],
                *[('X/2', f'Q{i}') for i in qubits],
                ('Barrier', tuple(f'Q{i}' for i in qubits)),
                *[(('Measure', j), f'Q{i}') for j, i in enumerate(qubits)],
            ], ],
            tuple([
                *[f'Q{i}' for i in qubits],
            ]): [
                sweep_list+kernel.get(f'gate.rfUnitary.Q{i}.params.beta') for i in qubits
            ],
        },
        'plot': [
            ['Beta', 'a.u.'], ['Probability' if signal ==
                               'state' else 'Amplitude', 'a.u.'],
        ],
    }


def ReadoutDelay_without_constrains(qubits: list[int], time_length: float, sweep_points: int = 41, signal: str = 'raw') -> dict:
    """
    ['time'] readout delay.

    Args:
        qubits (list[int]): the qubit id.
        time_length (float): delay time range, (-time_length, time_length) is applied.
        sweep_points (int, optional): sweep points. Defaults to 41.
        signal (str, optional): signal type. Defaults to 'raw'.

    Returns:
        dict: config.
    """

    assert time_length > 0, 'time_length should be positive.'

    delay_addition = {kernel.get(f'Q{i}.probe'): 0 for i in qubits}
    for j, i in enumerate(qubits):
        delay_addition[kernel.get(f'Q{i}.probe')] = max(delay_addition[kernel.get(
            f'Q{i}.probe')], kernel.get(f'gate.rfUnitary.Q{i}.params.duration')[1][-1])

    for j, i in enumerate(qubits):
        delay_addition[i] = delay_addition[kernel.get(
            f'Q{i}.probe')]-kernel.get(f'gate.rfUnitary.Q{i}.params.duration')[1][-1]

    return {
        'init': {
            'name': 'ReadoutDelay',
            'qubits': qubits,
            'signal': signal,
            'scanner_name': 'ReadoutDelay_without_constrains',
        },
        'setting': {
            'pre_setting': {},
            'compile_once': False,
            'circuit_type': 'cmds:qlispcmd',
            'circuit': [],
        },
        'constrains': [],
        'sweep_config': {
            'time': {
                'addr': None,
            },
            'circuit': {
                'addr': 'setting.circuit',
            },
        },
        'sweep_setting': {
            'time': np.linspace(-time_length, time_length, sweep_points),
        },
        'sweep_addition': {
            'circuit': lambda time: [
                *[(('Delay', time_length+delay_addition[i]),
                   f'Q{i}') for i in qubits],
                *[(('rfUnitary', np.pi, 0), f'Q{i}') for i in qubits],
                *[(('Delay', time), f'Q{i}') for i in qubits],
                ('Barrier', tuple(f'Q{i}' for i in qubits)),
                *[(('Measure', j), f'Q{i}') for j, i in enumerate(qubits)],
            ],
        },
        'plot': [
            ['Readout Delay', 's'], ['Probability' if signal ==
                                     'state' else 'Amplitude', 'a.u.'],
        ],
    }


def RTO_without_constrains(qubits: list[int], time_length: float, sweep_points: int = 5, signal: str = 'raw') -> dict:
    """
    [2, 'time'] In Ramsey experiment, x-axis projection and y-axis projection are measured replaced phase.

    Args:
        qubits (list[int]): the qubit id.
        time_length (float): length of delay time.
        sweep_points (int, optional): sweep points. Defaults to 5.
        signal (str, optional): signal type. Defaults to 'raw'.

    Returns:
        dict: config.
    """

    sweep_list = np.arange(0, time_length, time_length/sweep_points)

    return {
        'init': {
            'name': 'RTO',
            'qubits': qubits,
            'signal': signal,
            'scanner_name': 'RTO_without_constrains',
        },
        'setting': {
            'pre_setting': {},
            'compile_once': False,
            'circuit_type': 'cmds:qlispcmd',
            'circuit': [],
        },
        'constrains': [],
        'sweep_config': {
            'time': {
                'addr': None,
            },
            'circuit': {
                'addr': 'setting.circuit',
            },
            'gate': {
                'addr': None,
            },
        },
        'sweep_setting': {
            'gate': ['X/2', 'Y/2'],
            'time': sweep_list,
        },
        'sweep_addition': {
            'circuit': lambda time, gate: [
                *[('X/2', f'Q{i}') for i in qubits],
                *[(('Delay', time), f'Q{i}') for i in qubits],
                *[(gate, f'Q{i}') for i in qubits],
                ('Barrier', tuple(f'Q{i}' for i in qubits)),
                *[(('Measure', j), f'Q{i}') for j, i in enumerate(qubits)]
            ],
        },
        'plot': [
            ['Delay', 's'], ['Probability' if signal ==
                             'state' else 'Amplitude', 'a.u.'],
        ],
    }


def PiErrorOscillation_without_constrain(qubits: list[int], errorRate: float = 0.1, periods: int = 1, step: int = 1, signal: str = 'raw') -> dict:
    """
    ['count'] Multiply a known error on pi pulse to calibrate the error.

    Args:
        qubits (list[int]): the qubit id.
        errorRate (float, optional): error rate act on rfUnitary drive. Defaults to 0.1.
        periods (int, optional): number of period to measure. Defaults to 1.
        step (int, optional): 2*step size. Defaults to 1.
        signal (str, optional): signal type. Defaults to 'raw'.

    Returns:
        dict: config.
    """

    sweep_list = np.round(np.concatenate(
        (np.array([0]), np.arange(1, round(periods/errorRate), 2*step))))

    return {
        'init': {
            'name': 'PiErrorOscillation',
            'qubits': qubits,
            'signal': signal,
            'scanner_name': 'PiErrorOscillation_without_constrain',
        },
        'setting': {
            'pre_setting': {
                f'gate.rfUnitary.Q{i}.params.amp': [[0, 1], [0, (1-errorRate)*kernel.get(f'gate.rfUnitary.Q{i}.params.amp')[1][-1]]] for i in qubits
            },
            'compile_once': False,
            'circuit_type': 'cmds:qlispcmd',
            'circuit': [],
        },
        'constrains': [],
        'sweep_config': {
            'circuit': {
                'addr': 'setting.circuit',
            },
            'count': {
                'addr': None,
            },
        },
        'sweep_setting': {
            'count': sweep_list,
        },
        'sweep_addition': {
            'circuit': lambda count: [
                *[(('rfUnitary', np.pi, 0), f'Q{i}') for i in qubits]*count,
                ('Barrier', tuple(f'Q{i}' for i in qubits)),
                *[(('Measure', j), f'Q{i}') for j, i in enumerate(qubits)]
            ],
        },
        'plot': [
            ['Delay', 's'], ['Probability' if signal ==
                             'state' else 'Amplitude', 'a.u.'],
        ],
    }


def DRAGCheck_without_constrain(qubits: list[int], repeat: int = 1, signal: str = 'raw') -> dict:
    return {
        'init': {
            'name': 'DRAGCheck',
            'qubits': qubits,
            'signal': signal,
            'scanner_name': 'DRAGCheck_without_constrain',
            'shots': round(repeat)*1024,
        },
        'setting': {
            'pre_setting': {},
            'compile_once': False,
            'circuit_type': 'cmds:qlispcmd',
            'circuit': [],
        },
        'constrains': [],
        'sweep_config': {
            'circuit': {
                'addr': f'setting.circuit',
            },
        },
        'sweep_setting': {
            'circuit': [[
                *[(('Measure', j), f'Q{i}') for j, i in enumerate(qubits)],
            ], [
                *[('X', f'Q{i}') for i in qubits],
                *[('Y/2', f'Q{i}') for i in qubits],
                *[(('Measure', j), f'Q{i}') for j, i in enumerate(qubits)],
            ], [
                *[('Y', f'Q{i}') for i in qubits],
                *[('X/2', f'Q{i}') for i in qubits],
                *[(('Measure', j), f'Q{i}') for j, i in enumerate(qubits)],
            ], [
                *[('X', f'Q{i}') for i in qubits],
                *[(('Measure', j), f'Q{i}') for j, i in enumerate(qubits)],
            ], ],
        },
        'plot': [
            ['DRAG Scale', 'a.u.'], ['Probability' if signal ==
                                     'state' else 'Amplitude', 'a.u.'],
        ],
    }


def RB_single_qubit_without_constrain(qubits: list[int], rb_qubits: Optional[list[int]] = None, repeat: int = 1, signal: str = 'state',
                                      max_cycle: int = 500, base: Optional[list[list[tuple]]] = None, random_times: int = 20, start: int = 0,
                                      interleaves: Optional[list[str]] = None, sweep_points: int = 21) -> dict:
    """
    [None] RB on single qubit

    Args:
        qubits (list[int]): the readout qubit id
        rb_qubits (Optional[list[int]], optional): the qubit id which are acted on RB circuit. Defaults to None.
        repeat (int, optional): repeat times. Defaults to 1.
        signal (str, optional): signal type. Defaults to 'state'.
        max_cycle (int, optional): max cycle number. Defaults to 500.
        base (Optional[list[list[tuple]]], optional): base use. Defaults to None.
        random_times (int, optional): number of random times. Defaults to 20.
        interleaves (list[str], optional): interleaves gate. Defaults to [].
        sweep_points (int, optional): sweep points. Defaults to 21.

    Returns:
        dict: config.
    """

    from waveforms.quantum.rb import generateRBCircuit

    if rb_qubits is None:
        rb_qubits = qubits

    if interleaves is None:
        interleaves = []
    else:
        interleaves = [(interleave, 0) for interleave in interleaves]

    return {
        'init': {
            'name': 'RB_single_qubit',
            'qubits': qubits,
            'signal': signal,
            'scanner_name': 'RB_single_qubit_without_constrain',
            'shots': round(repeat)*1024,
        },
        'setting': {
            'pre_setting': {},
            'compile_once': False,
            'circuit_type': 'cmds:qlispcmd',
            'circuit': [],
        },
        'constrains': [],
        'sweep_config': {
            'circuit': {
                'addr': 'setting.circuit',
            },
            'cycle': {
                'addr': None,
            },
            'random_times': {
                'addr': None,
            }
        },
        'sweep_setting': {
            'random_times': np.arange(random_times),
            'cycle': generate_RB_cycle(max_cycle=max_cycle, sweep_points=sweep_points, start=start),
        },
        'sweep_addition': {
            'circuit': lambda cycle: [
                *[('X', f'Q{i}') for i in rb_qubits],
                *chain.from_iterable(generateRBCircuit(qubits=(f'Q{i}',), cycle=cycle,
                                                       seed=np.random.randint(0xfffffff), interleaves=interleaves, base=base) for i in rb_qubits),
                ('Barrier', tuple(f'Q{i}' for i in qubits)),
                *[(('Measure', j), f'Q{i}') for j, i in enumerate(qubits)],
            ],
        },
        'plot': [
            ['Count', 'a.u.'], ['Probability', 'a.u.'],
        ],
    }


def nop(qubits: list[int]) -> dict:
    """
    A null scanner.

    Args:
        qubits (list[int]): the qubit id

    Returns:
        dict: config.
    """

    return {
        'init': {
            'name': 'nop',
            'qubits': qubits,
            'scanner_name': 'nop',
        },
        'setting': {
            'pre_setting': {},
            'compile_once': False,
            'circuit_type': 'cmds:qlispcmd',
            'circuit': [],
        },
        'constrains': [],
        'sweep_config': {},
        'sweep_setting': {},
        'plot': [],
    }


def XEB_single_qubit_without_constrain(qubits: list[int], xeb_qubits: Optional[list[int]] = None, repeat: int = 1, signal: str = 'count', max_cycle: int = 40,
                                       random_times: int = 20, sweep_points: int = 21, start: float = 1) -> dict:
    """
    ['random_seed'] XEB on single qubit

    Args:
        qubits (list[int]): the readout qubit id
        xeb_qubits (Optional[list[int]], optional): the qubit id which are acted on XEB circuit. Defaults to None.
        repeat (int, optional): repeat times. Defaults to 1.
        signal (str, optional): signal type. Defaults to 'state'.
        max_cycle (int, optional): max cycle number. Defaults to 500.
        random_times (int, optional): number of random times. Defaults to 20.
        sweep_points (int, optional): sweep points. Defaults to 21.
        start (float, optional): 10^start to start in this measure to aviod too few cycles.

    Returns:
        dict: config.
    """

    from waveforms.quantum.xeb import generateXEBCircuit

    if xeb_qubits is None:
        xeb_qubits = qubits

    return {
        'init': {
            'name': 'XEB_single_qubit',
            'qubits': qubits,
            'signal': signal,
            'scanner_name': 'XEB_single_qubit_without_constrain',
            'shots': round(repeat)*1024,
        },
        'setting': {
            'pre_setting': {},
            'compile_once': False,
            'circuit_type': 'cmds:qlispcmd',
            'circuit': [],
        },
        'constrains': [],
        'sweep_config': {
            'circuit': {
                'addr': 'setting.circuit',
            },
            'cycle': {
                'addr': None,
            },
            'random_seed': {
                'addr': None,
            }
        },
        'sweep_setting': {
            'cycle': generate_RB_cycle(max_cycle=max_cycle, sweep_points=sweep_points, start=start),
            'random_seed': np.random.randint(0xfffffff, size=[random_times, ]),
        },
        'sweep_addition': {
            'circuit': lambda cycle, random_seed: [
                *generateXEBCircuit(qubits=xeb_qubits,
                                    cycle=cycle, seed=random_seed, mode='s'),
                ('Barrier', tuple(f'Q{i}' for i in qubits)),
                *[(('Measure', j), f'Q{i}') for j, i in enumerate(qubits)],
            ],
        },
        'plot': [
            ['Count', 'a.u.'], ['Probability', 'a.u.'],
        ],
    }


def XXDelta_without_constrain(qubits: list[int], n: int = 1, bound: float = 60e6, signal: str = 'state', sweep_points: int = 21, mode: str = 'linear') -> dict:
    """
    [f'Q{i}'] measure DRAG delta using a sequence as 'Y'-'X'-'-X'-'X'-'-X'-...-'X'-'-X'.

    Args:
        qubits (list[int]): the qubit id.
        n (int, optional): numbers of pulses 'X' and '-X'. Defaults to 1.
        bound (float, optional): delta upper bound, actually bound/n is applied. Defaults to 60e6.
        signal (str, optional): signal type. Defaults to 'state'.
        sweep_points (int, optional): sweep points. Defaults to 21.
        mode (str, optional): in ['linear' (for linear time sampling), 'log' (for log time sampling)]. Defaults to 'linear'.

    Returns:
        dict: config.
    """

    if mode == 'linear':
        sweep_list = generate_spanlist(center=0, delta=bound/n, sweep_points=sweep_points)
    elif mode == 'log':
        sweep_list = generate_spanlist_log(center=0, delta=bound/n, sweep_points=sweep_points)
    else:
        raise ValueError(f'Mode {mode} is not supported.')

    return {
        'init': {
            'name': 'X-XDelta',
            'qubits': qubits,
            'signal': signal,
            'scanner_name': 'XXDelta_without_constrain',
        },
        'setting': {
            'pre_setting': {},
            'compile_once': False,
            'circuit_type':
            'cmds:qlispcmd',
            'circuit': [
                *[(('rfUnitary', np.pi, np.pi/2), f'Q{i}') for i in qubits],
                *chain.from_iterable(zip([*[(('rfUnitary', np.pi, 0), f'Q{i}') for i in qubits] * n],
                                         [
                    *[(('rfUnitary', np.pi, np.pi), f'Q{i}')
                      for i in qubits] * n
                ])),
                ('Barrier', tuple(f'Q{i}' for i in qubits)),
                *[(('Measure', j), f'Q{i}') for j, i in enumerate(qubits)],
            ]
        },
        'constrains': [],
        'sweep_config': {
            f'Q{i}': {
                'addr': f'gate.rfUnitary.Q{i}.params.delta',
            } for i in qubits
        },
        'sweep_setting': {
            tuple([f'Q{i}' for i in qubits]): [
                sweep_list+kernel.get(f'gate.rfUnitary.Q{i}.params.delta') for i in qubits
            ],
        },
    }



# Author: lizhiyuan  <lizhiyuan@baqis.ac.cn>  2021/01/29


import numpy as np
import matplotlib.pyplot as plt


def T1(fit, ax=None):
    ax = plt.gca() if ax is None else ax
    ax.set_xlabel(r'Time ($\mu$s)', fontsize=5)
    ax.set_ylabel('Ampitude', fontsize=5)
    ax.set_title('Energy Relaxation', fontsize=5)
    ax.text(0.95, 0.95, '$T_1\ =\ %.1f(%.2f)\ \mu$s' % (fit.params['T1'], fit.error['T1']),
             horizontalalignment='right', verticalalignment='top', transform=ax.transAxes, fontsize=5)


def Rabi(fit, ax=None):
    ax = plt.gca() if ax is None else ax
    ax.set_xlabel(r'Time ($\mu$s)', fontsize=5)
    ax.set_ylabel('Ampitude', fontsize=5)
    ax.set_title('Rabi', fontsize=5)
    ax.text(0.95, 0.95, '''$T_r\ =\ %.1f\ \mu$s
                            $\Omega\ =\ %.3f$ MHz
                            $pplen\ =\ %.4f a.u.$''' % (
        fit.params['Tr'], fit.params['Omega'], fit.params['pplen']),
        horizontalalignment='right', verticalalignment='top', transform=ax.transAxes, fontsize=5)


def Ramsey(fit, ax=None):
    ax = plt.gca() if ax is None else ax
    ax.set_xlabel(r'Time ($\mu$s)', fontsize=5)
    ax.set_ylabel('Ampitude', fontsize=5)
    ax.set_title('Ramsey', fontsize=5)
    ax.text(0.95, 0.95, '''$T_{1}\ =\ %.1f\ \mu$s
                            $T_{2}^{\star}\ =\ %.1f\ \mu$s
                            $T_{\phi}\ =\ %.1f(%.2f)\ \mu$s
                            $\Delta\ =\ %.4f$ MHz''' % (
        fit.params['T1'], fit.params['T2star'], fit.params['Tphi'], fit.error['Tphi'], fit.params['Delta']),
        horizontalalignment='right', verticalalignment='top', transform=ax.transAxes, fontsize=5)


def Ramsey_withBeat(fit, ax=None):
    ax = plt.gca() if ax is None else ax
    ax.set_xlabel(r'Time ($\mu$s)', fontsize=5)
    ax.set_ylabel('Ampitude', fontsize=5)
    ax.set_title('Ramsey_withBeat', fontsize=5)
    ax.text(0.95, 0.95, '''$T_{1}\ =\ %.1f\ \mu$s
                            $T_{2}^{\star}\ =\ %.1f\ \mu$s
                            $T_{\phi}\ =\ %.1f(%.2f)\ \mu$s
                            $\Delta\ =\ %.4f$ MHz
                            $\Delta_2\ =\ %.4f$ MHz''' % (
        fit.params['T1'], fit.params['T2star'], fit.params['Tphi'], fit.error['Tphi'], fit.params['Delta'], fit.params['Delta2']),
        horizontalalignment='right', verticalalignment='top', transform=ax.transAxes, fontsize=5)


def SpinEcho(fit, ax=None):
    ax = plt.gca() if ax is None else ax
    ax.set_xlabel(r'Time ($\mu$s)', fontsize=5)
    ax.set_ylabel('Ampitude', fontsize=5)
    ax.set_title('SpinEcho', fontsize=5)
    ax.text(0.95, 0.95, '''$T_{1}\ =\ %.1f\ \mu$s
                            $T_{2E}\ =\ %.1f\ \mu$s
                            $T_{\phi}\ =\ %.1f(%.2f)\ \mu$s
                            $\Delta\ =\ %.4f$ MHz''' % (
        fit.params['T1'], fit.params['T2E'], fit.params['Tphi'], fit.error['Tphi'], fit.params['Delta']),
        horizontalalignment='right', verticalalignment='top', transform=ax.transAxes, fontsize=5)


def RB(fit, ax=None):
    ax = plt.gca() if ax is None else ax
    ax.set_xlabel(r'Gate Number', fontsize=5)
    ax.set_ylabel('Ampitude', fontsize=5)
    ax.set_title('Randomized Benchmarking', fontsize=5)
    ax.text(0.95, 0.95, '''$p\ =\ %.4f(%.5f)$
                            $F\ =\ %.4f$''' % (
        fit.params['p'], fit.error['p'],fit.params['Fidelity']),
        horizontalalignment='right', verticalalignment='top', transform=ax.transAxes, fontsize=5)



# Author: Liu Pei  <liupei200546@163.com>  2021/08/14

from typing import Literal, Optional

import numpy as np


__all__ = [
    'get_SSE', 'get_MSE', 'get_RMSE', 'get_SSR', 'get_SST', 'get_R2',
]


def get_SSE(y: np.ndarray, y_fit: np.ndarray, weights: Optional[np.ndarray] = None) -> float:
    """
    Return sum of squares for residuals (SSE) of y and y_fit, $\sum (y-y_fit)**2$

    Args:
        y (np.array): real data
        y_fit (np.array): estimation or fitting of data
        weights (Optional[np.array], optional): the weights of data. Defaults to None.

    Returns:
        np.float64: sum of squares for residuals(SSE)
    """

    if weights is None:
        weights = np.ones_like(y)
    return np.sum((y-y_fit)**2*weights)


def get_MSE(y: np.ndarray, y_fit: np.ndarray, weights: Optional[np.ndarray] = None) -> float:
    """
    Return mean squared error (MSE) of y and y_fit, $\sum (y-y_fit)**2/N$

    Args:
        y (np.array): real data
        y_fit (np.array): estimation or fitting of data
        weights (Optional[np.array], optional): the weights of data. Defaults to None.

    Returns:
        np.float64: mean squared error (MSE)
    """

    if weights is None:
        weights = np.ones_like(y)
    return np.sum((y-y_fit)**2*weights)/y.size


def get_RMSE(y: np.ndarray, y_fit: np.ndarray, weights: Optional[np.ndarray] = None) -> float:
    """
    Return root-mean-square error (RMSE) of y and y_fit, $\sqrt{\sum (y-y_fit)**2/N}$

    Args:
        y (np.array): real data
        y_fit (np.array): estimation or fitting of data
        weights (Optional[np.array], optional): the weights of data. Defaults to None.

    Returns:
        np.float64: root-mean-square error (RMSE)
    """

    if weights is None:
        weights = np.ones_like(y)
    return np.sqrt(np.sum((y-y_fit)**2*weights)/y.size)


def get_SSR(y: np.ndarray, y_fit: np.ndarray, weights: Optional[np.ndarray] = None) -> float:
    """
    Return sum of squares of the regression (SSR) of y and y_fit, $\sum (\bar y-y_fit)**2/N$

    Args:
        y (np.array): real data
        y_fit (np.array): estimation or fitting of data
        weights (Optional[np.array], optional): the weights of data. Defaults to None.

    Returns:
        np.float64: sum of squares of the regression (SSR)
    """

    if weights is None:
        weights = np.ones_like(y)
    return np.sum((np.mean(y)-y_fit)**2*weights)


def get_SST(y: np.ndarray, weights: Optional[np.ndarray] = None) -> float:
    """
    Return total sum of square (SST) of y and y_fit, $\sum (\bar y-y)**2/N$

    Args:
        y (np.array): real data
        weights (Optional[np.array], optional): the weights of data. Defaults to None.

    Returns:
        np.float64: total sum of square (SST)
    """

    if weights is None:
        weights = np.ones_like(y)
    return np.sum((np.mean(y)-y)**2*weights)


def get_R2(y: np.ndarray, y_fit: np.ndarray, weights: Optional[np.ndarray] = None, p: int = 1,
           mode: Literal['adjust', 'original'] = 'adjust') -> float:
    """
    Return R-square, i.e. R2 or adjustR2 of y and y_fit, 1-SSE/SST=SSR/SST

    Args:
        y (np.array): real data
        y_fit (np.array): estimation or fitting of data
        weights (Optional[np.array], optional): the weights of data. Defaults to None.
        p (int, optional): the number of independent variable. Defaults to 1.
        mode (str, optional): in 'original' return R2, in 'adjust' return adjusted R2. Defaults to 'adjust'.

    Returns:
        np.float64: R2 or adjustR2
    """
    n = y.size

    if weights is None:
        weights = np.ones_like(y)

    SSE = np.sum((y-y_fit)**2*weights)
    SST = np.sum((np.mean(y)-y)**2*weights)

    assert mode in ['original', 'adjust'], 'this mode is not supported.'

    if mode in ['original']:
        return 1-SSE/SST

    return 1-SSE*(n-1)/SST/(n-p-1)



# Author: Liu Pei  <liupei200546@163.com>  2021/08/14

import kernel
from qos_tools.experiment.scanner import Scanner


def quark_backend(**kw):

    task = kernel.create_task(Scanner, args=(), kwds=kw['init'])
    task.init(**kw)
    _task = kernel.submit(task)

    return _task



# Author: Liu Pei  <liupei200546@163.com>  2021/08/14

from graphlib import TopologicalSorter
from typing import Any, Callable, Optional

from kernel.sched import App
from qos_tools.gatemap import circuit_convert
from waveforms.scan_iter import Begin, End


def topological_order_of_constrains(
        constrains_list: list[tuple]) -> tuple[dict, dict, dict]:
    """
    Return the topological order of constrains list and other messages for calculation.

    Args:
        constrains_list (list): list[(keys, func, goal)]

    Raises:
        ValueError: [description]

    Returns:
        tuple[dict, dict, dict]: {goal:constrains}, {key(hashable object): list}, {key(hashable object): order(int)}
    """
    graph_core = TopologicalSorter()

    node_set = set()
    goal_set = set()
    constrains_dict = {}

    # NOTE : 
    # here seems like to use the library sorter to sort the constrans_list
    # where ks must be a collection of stuff same type(hashable) as goal, that
    # are the predesessor of this node, the node signature is goal,
    # the constrains_list is composed in the form like 
    #[
    #   (ks, f , goal)
    #   (ks, f , goal)
    #   ... 
    #   (ks, f , goal)
    #]
    # the first thing to returns is just A[goal] = (ks,f,goal)
    # the second return is the reversed edge direciton of A
    # the third is the order 
    # 
    # questions : 
    # 1. is this a closed graph. aka k in ks are in all the goals ? 
    # 2. topo sort did not make sure the key to be scanned to appear
    #   as close to the rear as possible 
    #   so that the repeatative tasks are done less.
    for keys, func, goal in constrains_list:
        if goal in goal_set:
            # the goal never repeat
            raise ValueError('This constrains is illegal.')
        else:
            goal_set.add(goal)
            node_set.add(goal)
            for ks in keys:
                node_set.add(ks) ; 
            # both the ks and goal is added into the node_set 
            graph_core.add(goal, *keys);# builidng the constraint_list into sorted graph
            constrains_dict[goal] = (keys, func, goal)

    # cn is the closed graph in this descripion.
    constrains_next = {item: [] for item in node_set}
    #{
    #  closed_goal :[]
    #}

    # reversing direction
    for keys, func, goal in constrains_list:
        for k in keys:
            constrains_next[k].append(goal)
    #{
    #   "w" :[ "n" that  "w"  depends on ] # contraints_dict like
    #   
    #   "w" :[ "n" that  depend on "w"] # constraints_next like
    #}
    try:
        order = graph_core.static_order()
        order_dict = {item: i for i, item in enumerate(order)} ;
        return constrains_dict, constrains_next, order_dict
    except Exception as e:
        # only exception should be the loop
        print(e)
        # why return if there is error ? 
        return {}, {}, {}


class Scanner(App):
    """
    Basic class for Scanner class. Change a config of scanning into a App(Task) object. Support the constrans updating.
    """
    #NOTE : This code is so bloated.
    def __init__(self,
                 name: str,
                 qubits: list[int],
                 scanner_name: str = '',
                 **kw):
        """
        Init.

        Args:
            name (str): the name of this experiment.
            qubits (list[int]): the qubits which are acted on.
            scanner_name (str): Defaults to None.
        """

        self._name = name
        self.qubits = qubits

        super().__init__(**kw)

    def init(self,
             sweep_config: dict = {},  # the configuration battery
             sweep_setting: dict = {}, # here defines what is to be scanned
             setting: dict = {},      # feed,feedback skip...
             constrains: list = [],   # the things to be used in topo sort
             sweep_addition: dict = {}, # why not just add this to the battery ? 
             sweep_trackers: list = [], # to take note on additional data
             sweep_filter: Optional[Callable] = None, 
             **kw): # this kws are for this Scanner, but not fore the App
        """
        Init.

        Args:
            sweep_config (dict, optional): main for sweep addr. Defaults to {}.
            sweep_setting (dict, optional): sweep list for scan_iters. Defaults to {}.
            setting (dict, optional): setting for this scanner. Defaults to {}.
            constrains (list, optional): constrains for this scanner. Defaults to [].
            sweep_addition (dict, optional): addition message for scan_iters. Defaults to {}.
            sweep_trackers (list, optional): list of tracker objclass. Defaults to [].
            sweep_filter (Optional[Callable], optional): filter in scan_iters. Defaults to None.
        """

        # recording without copy
        self.sweep_config = sweep_config
        self.sweep_setting = sweep_setting
        self.sweep_addition = sweep_addition
        self.sweep_trackers = sweep_trackers
        self.sweep_filter = sweep_filter

        self.setting = setting
        # adding a key
        self.setting['skip'] = 0
        
        # sort
        self.constrains, self.constrains_next, self.constrains_order = topological_order_of_constrains(
            constrains)

        
        if 'feedback' in setting.keys():
            self._cache = []
            assert setting.get(
                'level_marker', False), '`level_marker` must be True.'
            assert isinstance(setting['feedback'],
                              dict), '`feedback` must be a dict.'
            
            #XXX usually it is  (key,value) for dict.items.
            for value, key in setting['feedback'].items():
                assert isinstance(value, int) and isinstance(
                    key, Callable), 'Form of `feedback` is invalid.'
        # setting['feedback'] is like 
        # { 
        #   5 : f , 
        # } 

        if 'feed' in setting.keys():
            assert isinstance(
                setting['feed'], Callable), 'Method for feed of tracker does not exist.'
        # setting['feed'] is like : [ f ] 

    # hieracial dictionary with . as seperation
    # NOTE : did not iterate through list though.
    def get(self, key: str) -> Any:
        """
        get from self.setting and cached config. 

        Args:
            key (str): str, `.setting` allowed query as attribute.
            task ([type], optional): ScannerNode use. Defaults to None.

        Raises:
            e: failed in fetching keys

        Returns:
            Any: value
        """
        ks = key.split('.') if isinstance(key, str) else key
        if ks[0] in ['setting']:
            _v = self.setting
            for k in ks[1:]:
                # rec sink point
                # NOTE : cant the bottom types all just overide __getitem__ and __setitem__?
                try:
                    _v = _v[k]
                except TypeError:
                    _v = _v.get(k) 
                except Exception as e:
                    raise e
            return _v
        else:
            return super().get(key)

    def set(self, key: str, value: Any):
        """
        set in self.setting and cached config. 

        Args:
            key (str): `.setting` allowed query as attribute.
            value (Any): value. 
            task ([type], optional): ScannerNode use. Defaults to None.

        Raises:
            e: failed in fetching keys
        """

        if key is None:
            return

        ks = key.split('.') if isinstance(key, str) else key
        if ks[0] in ['setting']:
            if len(ks) > 1:
                _ks, _k = ks[:-1], ks[-1]
                _ct = self.get(_ks); # recursive sink point
            else:
                _k = key
                _ct = self.setting;
            #NOTE : should create the key path if not exist.
            assert _ct is not None, f"key '{key}' not exist !" 

            if value == _ct.get(_k):
                pass
            else:
                try:
                    _ct[_k] = value
                except TypeError:
                    # why is this a type error?
                    _ct.set(_k, value)
                except Exception as e:
                    # this action is not nessesary
                    raise e
        else:
            # letting the app do it.
            super().set(key, value, False)

    def update(self, update_dict: dict):
        """
        Update parameters under constraint check.

        Args:
            update_dict (dict): {key:value} to be updated.
            task ([type], optional): ScannerNode use. Defaults to None.

        Raises:
            ValueError: There are contradictions in the update, please check the constraint relationship.

        Returns:
            int: the flag whether self.skip is updated in this updating.
        """

        to_update_flag_dict = {
            item: False
            for item in self.constrains_order.keys()
        } # the check table 

      
        for keys in self.constrains_order.keys(): 
            # the "key" not the "keys" is integer order index
            if keys in update_dict: 
                value = update_dict[keys]
                if to_update_flag_dict[keys]:
                    ks = [
                        self.get(k)
                        for k in self.constrains[keys][0]
                    ]
                    if value != self.constrains[keys][1](*ks):
                        raise ValueError(
                            'There are contradictions in the update, please check the constraint relationship.'
                        )
                    else:
                        self.set(keys, value)
                else:
                    if keys in self.constrains.keys():
                        print(
                            'Warning! The relationship before the constraint is not checked'
                        )
                    self.set(keys, value)
                    for ks in self.constrains_next[keys]:
                        to_update_flag_dict[ks] = True
            elif to_update_flag_dict[keys]:
                ks = [self.get(k) for k in self.constrains[keys][0]]
                self.set(keys, self.constrains[keys][1](*ks))
                for ks in self.constrains_next[keys]:
                    to_update_flag_dict[ks] = True

        for keys, value in update_dict.items():
            if keys not in self.constrains_next.keys():
                self.set(keys, value)

    @property
    def name(self) -> str:
        return self._name

    def scan_range(self) -> dict:
        return {
            'loops': self.sweep_setting,
            'filter': self.sweep_filter,
            'functions': self.sweep_addition,
            'trackers': self.sweep_trackers,
            'level_marker': self.setting.get('level_marker', False)
        }

    def main(self):
        """
        Main process in a linear scanner.

        Args:
            task ([type], optional): ScannerNode use. Defaults to None.
        """
        self.update(self.setting.get('pre_setting', {}))

        last_skip = 0

        circuit = self.setting.get('circuit', [])
        convert_flag = self.setting.get('circuit_type', 'cmds:qlispcmd') != 'cmds:qlispcmd'
        feed_func = self.setting.get('feed', None)

        _cache_flag = self.setting.get('level_marker', False)

        for step in self.scan():

            if isinstance(step, Begin):
                continue
            elif isinstance(step, End):
                if step.level in self.setting['feedback'].keys():
                    self.flush()
                    self.setting['feedback'][step.level](self, step)

            else:
                skip = step.kwds.get('skip', last_skip)
                circuit = step.kwds.get('circuit', circuit)
                if convert_flag:
                    circuit = circuit_convert(
                        circuit, fr=self.setting['circuit_type'], to='cmds:qlispcmd')
                self.update(
                    {
                        v_dict['addr']: step.kwds[k]
                        for k, v_dict in self.sweep_config.items()
                    })
                tmp = self.exec(circuit, skip_compile=((
                    skip > 0) and (last_skip != skip) and (not self.setting.get('compile_once', False))))
                if _cache_flag:
                    self._cache.append(tmp)
                last_skip = skip

                if feed_func is not None:
                    feed_func(self, step)



# Rebuild: Liu Pei  <liupei200546@163.com>  2022/03/29

from itertools import chain
from typing import Optional

import kernel
import numpy as np

from .tools import generate_spanlist


def Spectrum(qubits: list[str],
             delta: Optional[float] = None, st: Optional[float] = None, ed: Optional[float] = None, sweep_points: int = 101, mode: str = 'linear',
             signal: str = 'iq_avg', act_qubits: Optional[list[str]] = None,
             with_other_params: list[tuple] = [], **kw) -> dict:
    """
    [f'{q}'] spectrum.

    Args:
        qubits (list[str]): the qubit name.
        delta (Optional[float], optional): sweep span. Defaults to None.
        st (Optional[float], optional): sweep start. Defaults to None.
        ed (Optional[float], optional): sweep end. Defaults to None.
        sweep_points (int, optional): sweep points. Defaults to 101.
        mode (str, optional): sample mode. Defaults to 'linear'.
        act_qubits (Optional[list[str]], optional): act qubits. Defaults to None.
        signal (str, optional): signal. Defaults to 'iq_avg'.
        with_other_params (list, optional): other params. Defaults to [].
    """

    act_qubits = qubits if act_qubits is None else act_qubits
    sweep_list = generate_spanlist(
        center=0, st=st, ed=ed, delta=delta, sweep_points=sweep_points, mode=mode)

    return {
        'init': {
            'name': 'Spectrum',
            'qubits': qubits,
            'signal': signal,
        },
        'sweep_setting': {
            tuple(act_qubits): tuple([
                kernel.get(f'gate.rfUnitary.{q}.params.frequency')+sweep_list for q in act_qubits
            ]),
        },
        'sweep_addition': {
            'circuit': lambda **kw: [
                *[(('rfUnitary', np.pi/2, 0,
                    ('with', ('param:frequency', kw[q]), *with_other_params)), q) for q in act_qubits],
                ('Barrier', tuple(qubits)),
                *[(('Measure', j), q) for j, q in enumerate(qubits)],
            ],
        },
    }


def PowerRabi(qubits: list[str],
              scale_times: float = 4, sweep_points: int = 21,
              signal: str = 'iq_avg', act_qubits: Optional[list[str]] = None,
              with_other_params: list[tuple] = [], **kw) -> dict:
    """
    [f'{q}'] power Rabi.

    Args:
        qubits (list[str]): the qubit name.
        scale_times (float, optional): proportion of max scale in this measure to the original. Defaults to 4.
        sweep_points (int, optional): sweep points. Defaults to 21.
        signal (str, optional): signal. Defaults to 'iq_avg'.
        act_qubits (Optional[list[str]], optional): act qubits. Defaults to None.
        with_other_params (list[tuple], optional): other params. Defaults to [].
    """

    act_qubits = qubits if act_qubits is None else act_qubits

    return {
        'init': {
            'name': 'Power Rabi',
            'qubits': qubits,
            'signal': signal,
        },
        'sweep_setting': {
            tuple(act_qubits): tuple([
                np.linspace(0, min(1, scale_times*kernel.get(f'gate.rfUnitary.{q}.params.amp')[1][-1]), sweep_points) for q in act_qubits
            ]),
        },
        'sweep_addition': {
            'circuit': lambda **kw: [
                *[(('rfUnitary', np.pi, 0,
                    ('with', ('param:amp', [[0, 0.5], [0, kw[q]]]), *with_other_params)), q) for q in act_qubits],
                ('Barrier', tuple(qubits)),
                *[(('Measure', j), q) for j, q in enumerate(qubits)],
            ],
        },
    }


def PowerRabi_n(qubits: list[str],
                n: int = 1, mode: str = 'log', sweep_points: int = 21,
                signal: str = 'iq_avg', act_qubits: Optional[list[str]] = None,
                with_other_params: list[tuple] = [], **kw) -> dict:
    """
    [f'{q}'] n pulses of power Rabi.

    Args:
        qubits (list[str]): the qubit name.
        n (int, optional): numbers of pulses. Defaults to 1.
        mode (str, optional): sample mode. Defaults to 'log'.
        sweep_points (int, optional): sweep points. Defaults to 21.
        signal (str, optional): signal. Defaults to 'iq_avg'.
        act_qubits (Optional[list[str]], optional): act qubits. Defaults to None.
        with_other_params (list[tuple], optional): other params. Defaults to [].
    """

    act_qubits = qubits if act_qubits is None else act_qubits
    scale = {q: kernel.get(f'gate.rfUnitary.{q}.params.amp')
             [-1][-1] for q in qubits}

    return {
        'init': {
            'name': 'N Power Rabi',
            'qubits': qubits,
            'signal': signal,
        },
        'sweep_setting': {
            tuple(act_qubits): tuple([
                generate_spanlist(center=scale[q], st=scale[q]*(1-1/n), ed=min(1, scale[q]*(1+1/n)), sweep_points=sweep_points, mode=mode) for q in act_qubits
            ]),
        },
        'sweep_addition': {
            'circuit': lambda n=n, **kw: [
                *[(('rfUnitary', np.pi, 0, ('with',
                                            ('param:amp', [[0, 0.5], [0, kw[q]]]), *with_other_params)), q) for q in act_qubits]*n,
                ('Barrier', tuple(qubits)),
                *[(('Measure', j), q) for j, q in enumerate(qubits)],
            ],
        },
    }


def TimeRabi(qubits: list[str],
             duration_times: float = 4, sweep_points: int = 21,
             signal: str = 'iq_avg', act_qubits: Optional[list[str]] = None,
             with_other_params: list[tuple] = [], **kw) -> dict:
    """
    [f'{q}'] time Rabi.

    Args:
        qubits (list[str]): the qubit name.
        duration_times (float, optional): proportion of max duration in this measure to the original. Defaults to 4.
        sweep_points (int, optional): sweep points. Defaults to 21.
        signal (str, optional): signal. Defaults to 'iq_avg'.
        act_qubits (Optional[list[str]], optional): act qubits. Defaults to None.
        with_other_params (list[tuple], optional): other params. Defaults to [].
    """

    act_qubits = qubits if act_qubits is None else act_qubits

    return {
        'init': {
            'name': 'Time Rabi',
            'qubits': qubits,
            'signal': signal,
        },
        'sweep_setting': {
            tuple(act_qubits): tuple([
                np.linspace(0, duration_times*kernel.get(f'gate.rfUnitary.{q}.params.duration')[1][-1], sweep_points) for q in act_qubits
            ]),
        },
        'sweep_addition': {
            'circuit': lambda **kw: [
                *[(('rfUnitary', np.pi, 0,
                    ('with', ('param:duration', [[0, 0.5], [kw[q], kw[q]]]), *with_other_params)), q) for q in act_qubits],
                ('Barrier', tuple(qubits)),
                *[(('Measure', j), q) for j, q in enumerate(qubits)],
            ],
        },
    }


def XXDelta(qubits: list[str],
            n: int = 1, bound: float = 60e6, mode: str = 'linear', sweep_points: int = 21, ini_gate=('rfUnitary', np.pi, np.pi/2),
            signal: str = 'iq_avg', act_qubits: Optional[list[str]] = None,
            with_other_params: list[tuple] = [], **kw) -> dict:
    """
    [f'{q}'] measure DRAG delta using a sequence as 'Y'-'X'-'-X'-'X'-'-X'-...-'X'-'-X'.

    Args:
        qubits (list[str]): the qubit name.
        n (int, optional): numbers of pulses 'X' and '-X'. Defaults to 1.
        bound (float, optional): delta upper bound, actually bound/n is applied. Defaults to 60e6.
        mode (str, optional): in ['linear' (for linear time sampling), 'log' (for log time sampling)]. Defaults to 'linear'.
        sweep_points (int, optional): sweep points. Defaults to 21.
        ini_gate (tuple, optional): first gate. Defaults to ('rfUnitary', np.pi, np.pi/2).
        signal (str, optional): signal. Defaults to 'iq_avg'.
        act_qubits (Optional[list[str]], optional): act qubits. Defaults to None.
        with_other_params (list[tuple], optional): other params. Defaults to [].
    """

    act_qubits = qubits if act_qubits is None else act_qubits
    sweep_list = generate_spanlist(
        center=0, delta=bound/n, sweep_points=sweep_points, mode=mode)

    return {
        'init': {
            'name': 'X-X Delta',
            'qubits': qubits,
            'signal': signal,
        },
        'sweep_setting': {
            tuple(act_qubits): tuple([
                sweep_list+kernel.get(f'gate.rfUnitary.{q}.params.delta') for q in act_qubits
            ]),
        },
        'sweep_addition': {
            'circuit': lambda n=n, **kw: [
                *[(ini_gate, q) for q in act_qubits],
                *chain.from_iterable(zip([*[(('rfUnitary', np.pi, 0,
                                              ('with', ('param:delta', kw[q]), *with_other_params)), q) for q in act_qubits]*n],
                                         [*[(('rfUnitary', np.pi, np.pi,
                                              ('with', ('param:delta', kw[q]), *with_other_params)), q) for q in act_qubits]*n])),
                ('Barrier', tuple(qubits)),
                *[(('Measure', j), q) for j, q in enumerate(qubits)],
            ],
        },
    }


def PowerRabi_n_buffer(qubits: list[str], buffer_list: np.ndarray,
                       n: int = 1, mode: str = 'log', sweep_points: int = 21,
                       signal: str = 'iq_avg', act_qubits: Optional[list[str]] = None,
                       with_other_params: list[tuple] = [], **kw) -> dict:
    """
    [f'{q}'] n pulses of power Rabi.

    Args:
        qubits (list[str]): the qubit name.
        n (int, optional): numbers of pulses. Defaults to 1.
        mode (str, optional): sample mode. Defaults to 'log'.
        sweep_points (int, optional): sweep points. Defaults to 21.
        signal (str, optional): signal. Defaults to 'iq_avg'.
        act_qubits (Optional[list[str]], optional): act qubits. Defaults to None.
        with_other_params (list[tuple], optional): other params. Defaults to [].
    """

    act_qubits = qubits if act_qubits is None else act_qubits
    scale = {q: kernel.get(f'gate.rfUnitary.{q}.params.amp')
             [-1][-1] for q in qubits}

    return {
        'init': {
            'name': 'N Power Rabi Buffer',
            'qubits': qubits,
            'signal': signal,
        },
        'sweep_setting': {
            tuple(act_qubits): tuple([
                generate_spanlist(center=scale[q], st=scale[q]*(1-1/n), ed=min(1, scale[q]*(1+1/n)), sweep_points=sweep_points, mode=mode) for q in act_qubits
            ]),
            'buffer': buffer_list,
        },
        'sweep_addition': {
            'circuit': lambda buffer, n=n, **kw: [
                *[(('rfUnitary', np.pi, 0,
                    ('with', ('param:amp', [[0, 0.5], [0, kw[q]]]), ('param:buffer', buffer), *with_other_params)), q) for q in act_qubits]*n,
                ('Barrier', tuple(qubits)),
                *[(('Measure', j), q) for j, q in enumerate(qubits)],
            ],
        },
    }


def XXDelta_buffer(qubits: list[str], buffer_list: np.ndarray,
                   n: int = 1, bound: float = 60e6, mode: str = 'linear', sweep_points: int = 21, ini_gate=('rfUnitary', np.pi, np.pi/2),
                   signal: str = 'iq_avg', act_qubits: Optional[list[str]] = None,
                   with_other_params: list[tuple] = [], **kw) -> dict:
    """
    [f'{q}'] measure DRAG delta using a sequence as 'Y'-'X'-'-X'-'X'-'-X'-...-'X'-'-X'.

    Args:
        qubits (list[str]): the qubit name.
        buffer_list (np.ndarray): array of buffer.
        n (int, optional): numbers of pulses 'X' and '-X'. Defaults to 1.
        bound (float, optional): delta upper bound, actually bound/n is applied. Defaults to 60e6.
        mode (str, optional): in ['linear' (for linear time sampling), 'log' (for log time sampling)]. Defaults to 'linear'.
        sweep_points (int, optional): sweep points. Defaults to 21.
        ini_gate (tuple, optional): first gate. Defaults to ('rfUnitary', np.pi, np.pi/2).
        signal (str, optional): signal. Defaults to 'iq_avg'.
        act_qubits (Optional[list[str]], optional): act qubits. Defaults to None.
        with_other_params (list[tuple], optional): other params. Defaults to [].
    """

    act_qubits = qubits if act_qubits is None else act_qubits
    sweep_list = generate_spanlist(
        center=0, delta=bound/n, sweep_points=sweep_points, mode=mode)

    return {
        'init': {
            'name': 'X-X Delta Buffer',
            'qubits': qubits,
            'signal': signal,
        },
        'sweep_config': {
            q: {
                'addr': f'gate.rfUnitary.{q}.params.delta',
            } for q in qubits
        },
        'sweep_setting': {
            tuple(act_qubits): tuple([
                sweep_list+kernel.get(f'gate.rfUnitary.{q}.params.delta') for q in act_qubits
            ]),
            'buffer': buffer_list,
        },
        'sweep_addition': {
            'circuit': lambda buffer, n=n, **kw: [
                *[(ini_gate, q) for q in act_qubits],
                *chain.from_iterable(zip([*[(('rfUnitary', np.pi, 0,
                                              ('with', ('param:buffer', buffer), ('param:delta', kw[q]), *with_other_params)), q) for q in act_qubits]*n],
                                         [*[(('rfUnitary', np.pi, np.pi,
                                              ('with', ('param:buffer', buffer), ('param:delta', kw[q]), *with_other_params)), q) for q in act_qubits]*n])),
                ('Barrier', tuple(qubits)),
                *[(('Measure', j), q) for j, q in enumerate(qubits)],
            ],
        },
    }



# Author: lizhiyuan  <lizhiyuan@baqis.ac.cn>  2021/01/01

from contextlib import contextmanager
import numpy as np
import json
import functools

from ._cm import ConsManager


def json_default(obj):
    if isinstance(obj, complex):
        return {'real': obj.real, 'imag': obj.imag}
    elif isinstance(obj, np.ndarray):
        return obj.tolist()
    return obj


def json_hook(dic):
    if 'imag' in dic.keys():
        return complex(dic['real'], dic['imag'])
    return dic


def get_dumps_func(backend='json', **kw):
    if backend in ['json']:
        default = dict(sort_keys=False, indent=2, default=json_default)
        default.update(**kw)
        return functools.partial(json.dumps, **default)
    else:
        raise Exception(f'{backend} is not support!')


def get_loads_func(backend='json', **kw):
    if backend in ['json']:
        return functools.partial(json.loads, object_hook=json_hook)
    else:
        raise Exception(f'{backend} is not support!')


def cfgload(fname='config', contrains=None, mode='json', **kw):
    file = f'{fname}.{mode}'
    loads_func = get_loads_func(mode, **kw)

    with open(file, 'r', encoding='utf8') as f:
        s = f.read()
    config = loads_func(s)

    if contrains is not None:
        if callable(contrains):
            _constrains = contrains(config)
        else:
            _constrains = contrains
        cfg = ConsManager(config, _constrains)
        config = cfg.todict()
    return config


def cfgsave(config, fname='config', contrains=None, mode='json', **kw):
    file = f'{fname}.{mode}'
    dumps_func = get_dumps_func(mode, **kw)

    if contrains is not None:
        if callable(contrains):
            _constrains = contrains(config)
        else:
            _constrains = contrains
        cfg = ConsManager(config, _constrains)
        config = cfg.todict()

    s = dumps_func(config)
    with open(file, 'w', encoding='utf8') as f:
        f.write(s)


@contextmanager
def cfgContext(fname='config', contrains=None, mode='json', **kw):
    config = cfgload(fname, contrains, mode, **kw)
    yield config
    cfgsave(config, fname, contrains, mode, **kw)



# Author: Liu Pei  <liupei200546@163.com>  2021/08/14

from graphlib import TopologicalSorter
from typing import Any, Callable, Optional

from kernel.sched import App
from qos_tools.gatemap import circuit_convert


def topological_order_of_constrains(constrains_list: list[tuple]) -> tuple[dict, dict, dict]:
    """
    Return the topological order of constrains list and other messages for calculation.

    Args:
        constrains_list (list): list[(keys, func, goal)]

    Raises:
        ValueError: [description]

    Returns:
        tuple[dict, dict, dict]: {goal:constrains}, {key(hashable object): list}, {key(hashable object): order(int)}
    """

    graph_core = TopologicalSorter()

    node_set = set()
    goal_set = set()
    constrains_dict = {}

    for keys, func, goal in constrains_list:
        if goal in goal_set:
            raise ValueError('This constrains is illegal.')
        else:
            goal_set.add(goal)
            node_set.add(goal)
            for ks in keys:
                node_set.add(ks)
            graph_core.add(goal, *keys)
            constrains_dict[goal] = (keys, func, goal)

    constrains_next = {item: [] for item in node_set}
    for keys, func, goal in constrains_list:
        for k in keys:
            constrains_next[k].append(goal)

    try:
        order = graph_core.static_order()
        order_dict = {item: i for i, item in enumerate(order)}
        return constrains_dict, constrains_next, order_dict
    except Exception as e:
        print(e)
        return {}, {}, {}


class Scanner(App):
    """
    Basic class for Scanner class. Change a config of scanning into a App(Task) object. Support the constrans updating.
    """

    def __init__(self, name: str, qubits: list[int], scanner_name: str = '', **kw):
        """
        Init.

        Args:
            name (str): the name of this experiment.
            qubits (list[int]): the qubits which are acted on.
            scanner_name (str): Defaults to None.
        """

        self._name = name
        self.qubits = qubits

        super().__init__(**kw)

        self.meta_info['scanner'] = scanner_name

    def init(self, sweep_config: dict, sweep_setting: dict, setting: dict, constrains: list = [],
             sweep_addition: dict = {}, mask_func: Optional[Callable] = None,
             plot_setting: Optional[dict] = None, plot: list[list[str]] = [['', ''], ['', '']], **kw):
        """
        Init.

        Args:
            sweep_config (dict): main for sweep addr.
            sweep_setting (dict): sweep list for scan_iters.
            setting (dict): setting for this scanner.
            constrains (list, optional): constrains for this scanner. Defaults to [].
            sweep_addition (dict, optional): addition message for scan_iters. Defaults to {}.
            mask_func (Optional[Callable], optional): filter in scan_iters. Defaults to None.
            plot_setting (Optional[dict], optional): use in UI. Defaults to None.
            plot (list[list[str]], optional): use in UI. Defaults to [['', ''], ['', '']].
        """

        self.sweep_config = sweep_config
        self.sweep_setting = sweep_setting
        self.sweep_addition = sweep_addition

        self.setting = setting
        self.setting['skip'] = 0
        self.constrains, self.constrains_next, self.constrains_order = topological_order_of_constrains(
            constrains)

        if isinstance(mask_func, str):
            mask_func = eval(mask_func)
        self.mask_func = mask_func

        self.plot_setting = plot_setting

        self.meta_info['name'] = self.name
        self.meta_info['axis'] = self.plot_setting

        self.meta_info['plot'] = plot

    def get(self, key: str, task=None) -> Any:
        """
        get from self.setting and cached config. 

        Args:
            key (str): str, `.setting` allowed query as attribute.
            task ([type], optional): ScannerNode use. Defaults to None.

        Raises:
            e: failed in fetching keys

        Returns:
            Any: value
        """
        ks = key.split('.') if isinstance(key, str) else key
        if ks[0] in ['setting']:
            _v = self.setting
            for k in ks[1:]:
                try:
                    _v = _v[k]
                except TypeError:
                    _v = _v.get(k)
                except Exception as e:
                    raise e
            # print(f'get key = {key}, value = {_v}')
            return _v
        else:
            # print(f'get key = {key}, value = {super().get(key)}')
            if task is not None:
                return task.get(key)
            else:
                return super().get(key)

    def set(self, key: str, value: Any, task=None) -> None:
        """
        set in self.setting and cached config. 

        Args:
            key (str): `.setting` allowed query as attribute.
            value (Any): value. 
            task ([type], optional): ScannerNode use. Defaults to None.

        Raises:
            e: failed in fetching keys
        """

        if key is None:
            return

        ks = key.split('.') if isinstance(key, str) else key
        if ks[0] in ['setting']:
            if len(ks) > 1:
                _ks, _k = ks[:-1], ks[-1]
                _ct = self.get(_ks)
            else:
                _k = key
                _ct = self.setting
            assert _ct is not None, f"key '{key}' not exist !"

            if value == _ct.get(_k):
                pass
            else:
                try:
                    _ct[_k] = value
                except TypeError:  # object is not subscriptable
                    _ct.set(_k, value)
                except Exception as e:
                    raise e
        else:
            # print(f'set key = {key}, value = {value}')
            if task is not None:
                task.set(key, value, False)
            else:
                super().set(key, value, False)

    def update(self, update_dict: dict, task=None) -> int:
        """
        Update parameters under constraint check.

        Args:
            update_dict (dict): {key:value} to be updated.
            task ([type], optional): ScannerNode use. Defaults to None.

        Raises:
            ValueError: There are contradictions in the update, please check the constraint relationship.

        Returns:
            int: the flag whether self.skip is updated in this updating.
        """

        last_skip = self.setting['skip']+0

        to_update_flag_dict = {
            item: False for item in self.constrains_order.keys()}

        for keys in self.constrains_order.keys():

            if keys in update_dict.keys():
                value = update_dict[keys]
                if to_update_flag_dict[keys]:
                    ks = [self.get(k, task=task)
                          for k in self.constrains[keys][0]]
                    if value != self.constrains[keys][1](*ks):
                        raise ValueError(
                            'There are contradictions in the update, please check the constraint relationship.')
                    else:
                        self.set(keys, value, task=task)
                else:
                    if keys in self.constrains.keys():
                        print(
                            'Warning! The relationship before the constraint is not checked')
                    self.set(keys, value, task=task)
                    for ks in self.constrains_next[keys]:
                        to_update_flag_dict[ks] = True
            elif to_update_flag_dict[keys]:
                ks = [self.get(k, task=task) for k in self.constrains[keys][0]]
                self.set(keys, self.constrains[keys][1](*ks), task=task)
                for ks in self.constrains_next[keys]:
                    to_update_flag_dict[ks] = True

        for keys, value in update_dict.items():
            if keys not in self.constrains_next.keys():
                self.set(keys, value, task=task)

        return self.setting['skip'] > 0 and last_skip != self.setting['skip']

    @property
    def name(self) -> str:
        return self._name

    def scan_range(self) -> dict:
        return {
            'iters': self.sweep_setting,
            'filter': self.mask_func,
            'additional_kwds': self.sweep_addition,
        }

    def main(self, task=None) -> None:
        """
        Main process in a linear scanner.

        Args:
            task ([type], optional): ScannerNode use. Defaults to None.
        """
        self.update(self.setting['pre_setting'], task=task)

        for step in self.scan() if task is None else task.scan():

            skip = self.update(
                {self.sweep_config[k]['addr']: v for k, v in step.kwds.items()}, task=task)
            circuit = circuit_convert(
                self.setting['circuit'], fr=self.setting['circuit_type'], to='cmds:qlispcmd')
            if task is None:
                self.exec(
                    circuit, skip_compile=skip and not self.setting['compile_once'])
            else:
                task.exec(
                    circuit, skip_compile=skip and not self.setting['compile_once'])
